@@ -17,38 +17,29 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 675 Mass Ave, Cambridge, MA 02139, USA.
  * 
- * $Id: daemon.c,v 1.60 2003/03/03 17:24:26 cheusov Exp $
+ * $Id: daemon.c,v 1.76 2004/05/15 16:51:46 cheusov Exp $
  * 
  */
 
 #include "dictd.h"
-#include <ctype.h>
-#include <setjmp.h>
-
+#include "index.h"
+#include "data.h"
 #include "md5.h"
 #include "regex.h"
 #include "dictdplugin.h"
 #include "strategy.h"
+#include "plugin.h"
 
-#ifndef HAVE_INET_ATON
-#define inet_aton(a,b) (b)->s_addr = inet_addr(a)
-#endif
-
-#ifndef HAVE_SNPRINTF
-extern int snprintf(char *str, size_t size, const char *format, ...);
-#endif
-
-#ifndef HAVE_VSNPRINTF
-extern int vsnprintf(char *str, size_t size, const char *format, va_list ap);
-#endif
-
+#include <ctype.h>
+#include <setjmp.h>
 
 static int          _dict_defines, _dict_matches;
-static int          daemonS;
-static const char   *daemonHostname;
-static const char   *daemonIP;
-static int          daemonPort;
-static char         daemonStamp[256];
+static int          daemonS_in  = 0;
+static int          daemonS_out = 0;
+static const char   *daemonHostname  = NULL;
+static const char   *daemonIP        = NULL;
+static int          daemonPort       = -1;
+static char         daemonStamp[256] = "";
 static jmp_buf      env;
 static int          daemonMime;
 
@@ -326,7 +317,9 @@ static int daemon_check_list( const char *user, lst_List acl )
    dictAccess   *a;
    int          retcode;
 
-   if (!acl) return DICT_ALLOW;
+   if (!acl)
+      return DICT_ALLOW;
+
    for (p = lst_init_position(acl); p; p = lst_next_position(p)) {
       a = lst_get_position(p);
       switch (a->type) {
@@ -388,7 +381,8 @@ void daemon_terminate( int sig, const char *name )
 {
    alarm(0);
    tim_stop( "t" );
-   close(daemonS);
+   close(daemonS_in);
+   close(daemonS_out);
    if (name) {
       daemon_log( DICT_LOG_TERM,
 		  "%s: d/m/c = %d/%d/%d; %sr %su %ss\n",
@@ -410,9 +404,11 @@ void daemon_terminate( int sig, const char *name )
                   dict_format_time( tim_get_user( "t" ) ),
                   dict_format_time( tim_get_system( "t" ) ) );
    }
+
    log_close();
    longjmp(env,1);
    if (sig) exit(sig+128);
+
    exit(0);
 }
 
@@ -423,7 +419,7 @@ static void daemon_write( const char *buf, int len )
    int count;
    
    while (left) {
-      if ((count = write(daemonS, buf, left)) != left) {
+      if ((count = write(daemonS_out, buf, left)) != left) {
 	 if (count <= 0) {
 	    if (errno == EPIPE) {
 	       daemon_terminate( 0, "pipe" );
@@ -462,6 +458,10 @@ static void daemon_crlf( char *d, const char *s, int dot )
       }
    }
    if (dot) {                   /* add final . */
+      if (!first){
+	 *d++ = '\r';
+	 *d++ = '\n';
+      }
       *d++ = '.';
       *d++ = '\r';
       *d++ = '\n';
@@ -484,7 +484,7 @@ static void daemon_printf( const char *format, ... )
       daemon_terminate( 0, __FUNCTION__ );
    }
 
-   pt = alloca(2*len);
+   pt = alloca(2*len + 10); /* +10 for the case when buf == "\n"*/
    daemon_crlf(pt, buf, 0);
    daemon_write(pt, strlen(pt));
 }
@@ -504,7 +504,7 @@ static void daemon_text( const char *text )
 
 static int daemon_read( char *buf, int count )
 {
-   return net_read( daemonS, buf, count );
+   return net_read( daemonS_in, buf, count );
 }
 
 static void daemon_ok( int code, const char *string, const char *timer )
@@ -575,6 +575,7 @@ static void daemon_dump_defs( lst_List list )
    unsigned long previousStart = 0;
    unsigned long previousEnd   = 0;
    const char *  previousDef   = NULL;
+   int count;
 
    LST_ITERATE(list,p,dw) {
       db = dw->database;
@@ -605,6 +606,16 @@ static void daemon_dump_defs( lst_List list )
 	  db->invisible ? ""  : db->databaseShort);
 
       daemon_mime();
+
+      if (db->filter){
+	 count = strlen(buf);
+	 daemon_log( DICT_LOG_AUTH, "filtering with: %s\ncount: %d\n",
+		     db->filter, count );
+
+	 dict_data_filter(buf, &count, strlen (buf), db->filter);
+	 buf[count] = '\0';
+      }
+
       daemon_text(buf);
       xfree( buf );
    }
@@ -668,9 +679,9 @@ static void daemon_banner( void )
 
    snprintf( daemonStamp, sizeof (daemonStamp), "<%d.%d.%lu@%s>", 
 	    _dict_forks,
-	    getpid(),
+	    (int) getpid(),
 	    (long unsigned)t,
-	    net_hostname() );
+	     net_hostname() );
    daemon_printf( "%d %s %s <auth.mime> %s\n",
 		  CODE_HELLO,
                   net_hostname(),
@@ -721,17 +732,26 @@ static void daemon_define( const char *cmdline, int argc, char **argv )
 		     actual_matches );
       daemon_dump_defs( list );
       daemon_ok( CODE_OK, "ok", "c" );
+#ifdef USE_PLUGIN
+      call_dictdb_free (DictConfig->dbl);
+#endif
       dict_destroy_list( list );
       return;
    }
 
    if (!db_found) {
+#ifdef USE_PLUGIN
+      call_dictdb_free (DictConfig->dbl);
+#endif
       dict_destroy_list( list );
       daemon_printf( "%d invalid database, use SHOW DB for list\n",
 		     CODE_INVALID_DB );
       return;
    }
 
+#ifdef USE_PLUGIN
+   call_dictdb_free (DictConfig->dbl);
+#endif
    dict_destroy_list( list );
    daemon_log( DICT_LOG_NOMATCH,
 	       "%s exact \"%s\"\n", databaseName, word );
@@ -791,17 +811,26 @@ static void daemon_match( const char *cmdline, int argc, char **argv )
 		     CODE_MATCHES_FOUND, actual_matches );
       daemon_dump_matches( list );
       daemon_ok( CODE_OK, "ok", "c" );
+#ifdef USE_PLUGIN
+      call_dictdb_free (DictConfig->dbl);
+#endif
       dict_destroy_list( list );
       return;
    }
 
    if (!db_found) {
+#ifdef USE_PLUGIN
+      call_dictdb_free (DictConfig->dbl);
+#endif
       dict_destroy_list( list );
       daemon_printf( "%d invalid database, use SHOW DB for list\n",
 		     CODE_INVALID_DB );
       return;
    }
 
+#ifdef USE_PLUGIN
+   call_dictdb_free (DictConfig->dbl);
+#endif
    dict_destroy_list( list );
    daemon_log( DICT_LOG_NOMATCH,
 	       "%s %s \"%s\"\n", databaseName, strategy, word );
@@ -828,21 +857,28 @@ static dictDatabase *next_database (
       if (*databasePosition) {
 	 do {
 	    db = lst_get_position( *databasePosition );
+
 	    *databasePosition = lst_next_position( *databasePosition );
-	 } while (db && !db->available);
+	 } while ( db && (!db->available || db->invisible));
       }
       return db;
    } else {
       while (*databasePosition) {
          db = lst_get_position( *databasePosition );
 	 *databasePosition = lst_next_position( *databasePosition );
-         if (db && !strcmp(db -> databaseName,name) && !db -> invisible) {
-	    if (db->available)
+
+         if (db){
+	    if (
+	       !db -> invisible && db->available &&
+	       !strcmp(db -> databaseName,name))
+	    {
 	       return db;
-	    else
-               return NULL;
+	    }
+	 }else{
+	    return NULL;
 	 }
       }
+
       return NULL;
    }
 }
@@ -855,9 +891,9 @@ static int count_databases( void )
    lst_Position databasePosition = first_database_pos ();
 
    while (NULL != (db = next_database (&databasePosition, "*"))){
-      if (!db->invisible){
-	 ++count;
-      }
+      assert (!db -> invisible);
+
+      ++count;
    }
 
    return count;
@@ -875,26 +911,6 @@ static void destroy_word_list (lst_List l)
    lst_destroy (l);
 }
 
-/*
-  Replaces invisible databases with db argument.
- */
-static void replace_invisible_databases (
-   lst_Position *pos,
-   const dictDatabase *db)
-{
-   dictWord *dw;
-
-   while (pos){
-      dw = (dictWord *) lst_get_position (pos);
-
-      if (dw -> database && dw -> database -> invisible){
-	 dw -> database_visible = db;
-      }
-
-      pos = lst_next_position (pos);
-   }
-}
-
 /*
   Search for all words in word_list in the database db
  */
@@ -919,33 +935,15 @@ static int dict_search_words (
       word = lst_get_position (word_list_pos);
 
       if (word){
-	 if (db -> virtual_db_list){
-	    assert (lst_init_position (db -> virtual_db_list));
+	 matches_count = dict_search (
+	    l, word, db, strategy,
+	    result, extra_result, extra_result_size);
 
-	    old_count = lst_length (l);
+	 if (*result == DICT_PLUGIN_RESULT_PREPROCESS){
+	    assert (matches_count > 0);
 
-	    matches_count = dict_search_databases (
-	       l, lst_init_position (db -> virtual_db_list),
-	       "*", word, strategy, &db_found);
-
-	    assert (db_found);
-
-	    if (matches_count > 0){
-	       replace_invisible_databases (
-		  lst_nth_position (l, old_count + 1),
-		  db);
-	    }
-	 }else{
-	    matches_count = dict_search (
-	       l, word, db, strategy,
-	       result, extra_result, extra_result_size);
-
-	    if (*result == DICT_PLUGIN_RESULT_PREPROCESS){
-	       assert (matches_count > 0);
-
-	       xfree (lst_get_position (word_list_pos));
-	       lst_set_position (word_list_pos, NULL);
-	    }
+	    xfree (lst_get_position (word_list_pos));
+	    lst_set_position (word_list_pos, NULL);
 	 }
 
 	 if (matches_count < 0){
@@ -1074,10 +1072,11 @@ static void daemon_show_db( const char *cmdline, int argc, char **argv )
 
       daemon_mime();
       while ((db = next_database(&databasePosition, "*"))) {
-	 if (!db->invisible){
-	    daemon_printf( "%s \"%s\"\n",
-			   db->databaseName, db->databaseShort );
-	 }
+	 assert (!db->invisible);
+
+	 daemon_printf(
+	    "%s \"%s\"\n",
+	    db->databaseName, db->databaseShort );
       }
       daemon_printf( ".\n" );
       daemon_ok( CODE_OK, "ok", NULL );
@@ -1137,16 +1136,12 @@ static void daemon_show_info( const char *cmdline, int argc, char **argv )
    }
 
    list = lst_create();
-   while ((db = next_database(&databasePosition, argv[2] ))) {
-      if (
-	 db -> databaseInfo &&
-	 (db -> virtual_db || db -> databaseInfo [0] != '@'))
-      {
+   while ((db = next_database(&databasePosition, argv[2]))) {
+      if (db -> databaseInfo && db -> databaseInfo [0] != '@'){
 	 daemon_printf( "%d information for %s\n",
 			CODE_DATABASE_INFO, argv[2] );
 	 daemon_mime();
 	 daemon_text(db -> databaseInfo);
-	 daemon_text("\n");
 	 daemon_ok( CODE_OK, "ok", NULL );
 	 return;
       }
@@ -1164,6 +1159,10 @@ static void daemon_show_info( const char *cmdline, int argc, char **argv )
 	 dw = lst_nth_get( list, 1 );
 	 buf = dict_data_obtain( db, dw );
 
+#ifdef USE_PLUGIN
+	 call_dictdb_free (DictConfig->dbl);
+#endif
+
 	 dict_destroy_list( list );
 	 daemon_printf( "%d information for %s\n",
 			CODE_DATABASE_INFO, argv[2] );
@@ -1172,6 +1171,10 @@ static void daemon_show_info( const char *cmdline, int argc, char **argv )
 	 daemon_ok( CODE_OK, "ok", NULL );
 	 return;
       } else {
+#ifdef USE_PLUGIN
+	 call_dictdb_free (DictConfig->dbl);
+#endif
+
 	 dict_destroy_list( list );
 	 daemon_printf( "%d information for %s\n",
 			CODE_DATABASE_INFO, argv[2] );
@@ -1187,6 +1190,30 @@ static void daemon_show_info( const char *cmdline, int argc, char **argv )
 		  CODE_INVALID_DB );
 }
 
+static int daemon_get_max_dbname_length ()
+{
+   size_t max_len  = 0;
+   size_t curr_len = 0;
+
+   const dictDatabase *db;
+
+   lst_Position databasePosition = first_database_pos ();
+
+   while (NULL != (db = next_database (&databasePosition, "*"))){
+      assert (!db -> invisible);
+
+      if (db -> databaseName){
+	 curr_len = strlen (db -> databaseName);
+
+	 if (curr_len > max_len){
+	    max_len = curr_len;
+	 }
+      }
+   }
+
+   return (int) max_len;
+}
+
 static void daemon_show_server( const char *cmdline, int argc, char **argv )
 {
    FILE          *str;
@@ -1202,13 +1229,17 @@ static void daemon_show_server( const char *cmdline, int argc, char **argv )
    daemon_printf( "%d server information\n", CODE_SERVER_INFO );
    daemon_mime();
    daemon_printf( "%s\n", dict_get_banner(0) );
-   
-   daemon_printf( "On %s: up %s, %d fork%s (%0.1f/hour)\n",
-		  net_hostname(),
-		  dict_format_time( uptime ),
-		  _dict_forks,
-		  _dict_forks > 1 ? "s" : "",
-		  (_dict_forks/uptime)*3600.0 );
+
+   if (!inetd){
+      daemon_printf (
+	 "On %s: up %s, %d fork%s (%0.1f/hour)\n",
+	 net_hostname(),
+	 dict_format_time( uptime ),
+	 _dict_forks,
+	 _dict_forks > 1 ? "s" : "",
+	 (_dict_forks/uptime)*3600.0 );
+   }
+
    if (count_databases()) {
       daemon_printf( "\nDatabase      Headwords         Index"
 		     "          Data  Uncompressed\n" );
@@ -1223,8 +1254,9 @@ static void daemon_show_server( const char *cmdline, int argc, char **argv )
 	 int data_length     = 0;
 	 char data_length_uom= 'k';
 
-	 if (db -> invisible)
-	    continue;
+	 int max_dbname_len = 0;
+
+	 assert (!db -> invisible);
 
 	 if (db->index){
 	    index_size = db->index->size/1024 > 10240 ?
@@ -1242,8 +1274,13 @@ static void daemon_show_server( const char *cmdline, int argc, char **argv )
 	    data_length_uom = db->data->length/1024 > 10240 ? 'M' : 'k';
 	 }
 
+	 max_dbname_len = daemon_get_max_dbname_length ();
+
 	 daemon_printf(
-	    "%-12.12s %10lu %10lu %cB %10lu %cB %10lu %cB\n",
+	    "%-*.*s %10lu %10lu %cB %10lu %cB %10lu %cB\n",
+	    max_dbname_len,
+	    max_dbname_len,
+
 	    db->databaseName,
 	    headwords,
 
@@ -1253,8 +1290,8 @@ static void daemon_show_server( const char *cmdline, int argc, char **argv )
 	    data_size,
 	    data_size_uom,
 
-	    data_size,
-	    data_size_uom);
+	    data_length,
+	    data_length_uom);
       }
    }
 
@@ -1312,7 +1349,7 @@ static void daemon_auth( const char *cmdline, int argc, char **argv )
       daemon_printf( "%d syntax error, illegal parameters\n",
 		     CODE_ILLEGAL_PARAM );
    if (!h || !(secret = hsh_retrieve(h, argv[1]))) {
-      daemon_log( DICT_LOG_AUTH, "%s@%s/%s denied: no invalid username\n",
+      daemon_log( DICT_LOG_AUTH, "%s@%s/%s denied: invalid username\n",
                   argv[1], daemonHostname, daemonIP );
       daemon_printf( "%d auth denied\n", CODE_AUTH_DENIED );
       return;
@@ -1327,7 +1364,7 @@ static void daemon_auth( const char *cmdline, int argc, char **argv )
    MD5Final(digest, &ctx);
 
    for (i = 0; i < 16; i++)
-      snprintf( hex+2*i, 2, "%02x", digest[i] );
+      snprintf( hex+2*i, 3, "%02x", digest[i] );
 
    hex[32] = '\0';
 
@@ -1389,19 +1426,31 @@ static void daemon_quit( const char *cmdline, int argc, char **argv )
    daemon_terminate( 0, "quit" );
 }
 
+/* The whole sub should be moved here, but I want to keep the diff small. */
+int _handleconn (int delay, int error);
+
+int dict_inetd(char ***argv0, int delay, int error )
+{
+   if (setjmp(env)) return 0;
+
+   daemonPort = -1;
+   daemonIP   = "inetd";
+
+   daemonHostname = daemonIP;
+
+   daemonS_in        = 0;
+   daemonS_out       = 0;
+
+   return _handleconn(delay, error);
+}
+
 int dict_daemon( int s, struct sockaddr_in *csin, char ***argv0, int delay,
 		 int error )
 {
-   char           buf[4096];
-   int            count;
    struct hostent *h;
-   arg_List       cmdline;
-   int            argc;
-   char           **argv;
-   void           (*command)(const char *, int, char **);
-      
+	
    if (setjmp(env)) return 0;
-   
+
    daemonPort = ntohs(csin->sin_port);
    daemonIP   = str_find( inet_ntoa(csin->sin_addr) );
    if ((h = gethostbyaddr((void *)&csin->sin_addr,
@@ -1410,8 +1459,20 @@ int dict_daemon( int s, struct sockaddr_in *csin, char ***argv0, int delay,
    } else
       daemonHostname = daemonIP;
 
-   daemonS           = s;
-   
+   daemonS_in        = s;
+   daemonS_out       = s;
+
+   return _handleconn(delay, error);
+}
+
+int _handleconn (int delay, int error) {
+   char           buf[4096];
+   int            count;
+   arg_List       cmdline;
+   int            argc;
+   char           **argv;
+   void           (*command)(const char *, int, char **);
+      
    _dict_defines     = 0;
    _dict_matches     = 0;
    _dict_comparisons = 0;
@@ -1423,7 +1484,7 @@ int dict_daemon( int s, struct sockaddr_in *csin, char ***argv0, int delay,
    dict_setproctitle( "dictd: %s connected", daemonHostname );
 
    if (error) {
-      daemon_printf( "%d temporarily unavailable\n",
+      daemon_printf( "%d server temporarily unavailable\n",
 		     CODE_TEMPORARILY_UNAVAILABLE );
       daemon_terminate( 0, "temporarily unavailable" );
    }
