@@ -17,10 +17,13 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 675 Mass Ave, Cambridge, MA 02139, USA.
  * 
- * $Id: data.c,v 1.20 2003/03/04 12:31:38 cheusov Exp $
+ * $Id: data.c,v 1.23 2003/10/31 00:40:04 cheusov Exp $
  * 
  */
 
+#include "dictP.h"
+
+#include "data.h"
 #include "dictzip.h"
 
 #include <sys/stat.h>
@@ -128,7 +131,7 @@ static int dict_read_header( const char *filename,
       si1                  = getc( str );
       si2                  = getc( str );
       
-      if (si1 == GZ_RND_S1 || si2 == GZ_RND_S2) {
+      if (si1 == GZ_RND_S1 && si2 == GZ_RND_S2) {
 	 subLength            = getc( str ) << 0;
 	 subLength           |= getc( str ) << 8;
 	 header->version      = getc( str ) << 0;
@@ -162,8 +165,16 @@ static int dict_read_header( const char *filename,
    
    if (header->flags & GZ_FNAME) { /* FIXME! Add checking against header len */
       pt = buffer;
-      while ((c = getc( str )) && c != EOF)
+      while ((c = getc( str )) && c != EOF){
 	 *pt++ = c;
+
+	 if (pt == buffer + sizeof (buffer)){
+	    err_fatal (
+	       __FUNCTION__,
+	       "too long FNAME field in dzip file \"%s\"\n", filename);
+	 }
+      }
+
       *pt = '\0';
       header->origFilename = str_find( buffer );
       header->headerLength += strlen( header->origFilename ) + 1;
@@ -173,8 +184,16 @@ static int dict_read_header( const char *filename,
    
    if (header->flags & GZ_COMMENT) { /* FIXME! Add checking for header len */
       pt = buffer;
-      while ((c = getc( str )) && c != EOF)
+      while ((c = getc( str )) && c != EOF){
 	 *pt++ = c;
+
+	 if (pt == buffer + sizeof (buffer)){
+	    err_fatal (
+	       __FUNCTION__,
+	       "too long COMMENT field in dzip file \"%s\"\n", filename);
+	 }
+      }
+
       *pt = '\0';
       header->comment = str_find( buffer );
       header->headerLength += strlen( header->comment ) + 1;
@@ -209,7 +228,8 @@ static int dict_read_header( const char *filename,
 			      * header->chunkCount );
    for (offset = header->headerLength + 1, i = 0;
 	i < header->chunkCount;
-	i++) {
+	i++)
+   {
       header->offsets[i] = offset;
       offset += header->chunks[i];
    }
