@@ -17,11 +17,12 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 675 Mass Ave, Cambridge, MA 02139, USA.
  * 
- * $Id: dict.c,v 1.33 2003/02/02 16:34:26 hilliard Exp $
+ * $Id: dict.c,v 1.43 2004/05/24 11:42:25 cheusov Exp $
  * 
  */
 
 #include "dict.h"
+#include "parse.h"
 #include "md5.h"
 #include <stdarg.h>
 
@@ -72,6 +73,31 @@ unsigned long client_bytes;
 unsigned long client_pipesize = PIPESIZE;
 char          *client_text    = NULL;
 
+int ex_status = 0;
+static void set_ex_status (int status)
+{
+   if (!ex_status)
+      ex_status = status;
+}
+
+#define EXST_NO_MATCH                20
+#define EXST_APPROX_MATCHES          21
+#define EXST_NO_DATABASES            22
+#define EXST_NO_STRATEGIES           23
+
+#define EXST_UNEXPECTED              30
+#define EXST_TEMPORARILY_UNAVAILABLE 31
+#define EXST_SHUTTING_DOWN           32
+#define EXST_SYNTAX_ERROR            33
+#define EXST_ILLEGAL_PARAM           34
+#define EXST_COMMAND_NOT_IMPLEMENTED 35
+#define EXST_PARAM_NOT_IMPLEMENTED   36
+#define EXST_ACCESS_DENIED           37
+#define EXST_AUTH_DENIED             38
+#define EXST_INVALID_DB              39
+#define EXST_INVALID_STRATEGY        40
+#define EXST_CONNECTION_FAILED       41
+
 struct def {
    lst_List   data;
    const char *word;
@@ -124,6 +150,88 @@ static void client_crlf( char *d, const char *s )
    *d = '\0';
 }
 
+static void unexpected_status_code (
+   int expected_status, int act_status,
+   const char *message)
+{
+   switch (act_status){
+   case CODE_TEMPORARILY_UNAVAILABLE:
+      fprintf (stderr,
+	       "%s\n",
+	       (message ? message : "Server temporarily unavailable"));
+      exit (EXST_TEMPORARILY_UNAVAILABLE);
+
+   case CODE_SHUTTING_DOWN:
+      fprintf (stderr,
+	       "%s\n", (message ? message : "Server shutting down"));
+      exit (EXST_SHUTTING_DOWN);
+
+   case CODE_SYNTAX_ERROR:
+      fprintf (stderr,
+	       "%s\n", (message ? message : "Command not recognized"));
+      exit (EXST_SYNTAX_ERROR);
+
+   case CODE_ILLEGAL_PARAM:
+      fprintf (stderr,
+	       "%s\n", (message ? message : "Illegal parameters"));
+      exit (EXST_ILLEGAL_PARAM);
+
+   case CODE_COMMAND_NOT_IMPLEMENTED:
+      fprintf (stderr,
+	       "%s\n", (message ? message : "Command not implemented"));
+      exit (EXST_COMMAND_NOT_IMPLEMENTED);
+
+   case CODE_PARAM_NOT_IMPLEMENTED:
+      fprintf (stderr,
+	       "%s\n", (message ? message : "Parameter not implemented"));
+      exit (EXST_PARAM_NOT_IMPLEMENTED);
+
+   case CODE_ACCESS_DENIED:
+      fprintf (stderr,
+	       "%s\n", (message ? message : "Access denied"));
+      exit (EXST_ACCESS_DENIED);
+
+   case CODE_AUTH_DENIED:
+      fprintf (stderr,
+	       "%s\n", (message ? message : "Authentication denied"));
+      exit (EXST_AUTH_DENIED);
+
+   case CODE_INVALID_DB:
+      fprintf (stderr,
+	       "%s\n", (message ? message : "Invalid database"));
+      exit (EXST_INVALID_DB);
+
+   case CODE_INVALID_STRATEGY:
+      fprintf (stderr,
+	       "%s\n", (message ? message : "Invalid strategy"));
+      exit (EXST_INVALID_STRATEGY);
+
+   case CODE_NO_MATCH:
+      fprintf (stderr,
+	       "%s\n", (message ? message : "No matches found"));
+      exit (EXST_NO_MATCH);
+
+   case CODE_NO_DATABASES:
+      fprintf (stderr,
+	       "%s\n", (message ? message : "No databases"));
+      exit (EXST_NO_DATABASES);
+
+   case CODE_NO_STRATEGIES:
+      fprintf (stderr,
+	       "%s\n", (message ? message : "No strategies"));
+      exit (EXST_NO_STRATEGIES);
+
+   default:
+      fprintf (stderr,
+	       "Unexpected status code %d (%s), wanted %d\n",
+	       act_status,
+	       (message ? message : "no message"),
+	       expected_status);
+
+      exit (EXST_UNEXPECTED);
+   }
+}
+
 static void client_open_pager( void )
 {
    int infd;
@@ -210,6 +318,7 @@ static void client_print_matches( lst_List l, int flag, const char *word )
    } else {
        if (flag)
            fprintf( dict_output, "No matches found for \"%s\"\n", word );
+       set_ex_status (EXST_NO_MATCH);
        return;
    }
 
@@ -252,7 +361,7 @@ static void client_print_listed( lst_List l )
    arg_List     a;
    int          colWidth = 10; /* minimum size of first column */
    int          colMax = 16; /* maximum size of unragged first column */
-   char         format[BUFFERSIZE];
+   char         format[32];
 
    if (!l) return;
    LST_ITERATE(l,p,e) {
@@ -269,7 +378,7 @@ static void client_print_listed( lst_List l )
 
       arg_destroy(a);
    }
-   sprintf( format, " %%-%ds %%s\n", colWidth );
+   snprintf( format, sizeof (format), " %%-%ds %%s\n", colWidth );
    LST_ITERATE(l,p,e) {
       a = arg_argify( e, 0 );
       fprintf( dict_output, format, arg_get(a,0), arg_get(a,1) );
@@ -461,6 +570,7 @@ static void request( void )
       
    LST_ITERATE(cmd_list,pos,c) {
       b[0] = '\0';
+      len = 0;
       PRINTF(DBG_PIPE,("* Looking at request %d\n",c->command));
       if (c->sent) {
 	 PRINTF(DBG_PIPE,("* Skipping\n"));
@@ -478,32 +588,36 @@ static void request( void )
 	 MD5Update(&ctx, cmd_reply.msgid, strlen(cmd_reply.msgid));
 	 MD5Update(&ctx, cmd_reply.key, strlen(cmd_reply.key));
 	 MD5Final(digest, &ctx );
-	 for (i = 0; i < 16; i++) sprintf( hex+2*i, "%02x", digest[i] );
+	 for (i = 0; i < 16; i++)
+	    sprintf( hex+2*i, "%02x", digest[i] );
 	 hex[32] = '\0';
-	 sprintf( b, "auth %s %s\n", cmd_reply.user, hex );
+	 snprintf( b, BUFFERSIZE, "auth %s %s\n", cmd_reply.user, hex );
 	 break;
       case CMD_CLIENT:
          if (client_text)
-            sprintf( b, "client \"%s: %s\"\n", c->client, client_text );
+            snprintf( b, BUFFERSIZE, "client \"%s: %s\"\n",
+                            c->client, client_text );
          else
-            sprintf( b, "client \"%s\"\n", c->client );
+            snprintf( b, BUFFERSIZE, "client \"%s\"\n", c->client );
          break;
-      case CMD_INFO:   sprintf( b, "show info %s\n", c->database );   break;
-      case CMD_SERVER: sprintf( b, "show server\n" );                 break;
-      case CMD_DBS:    sprintf( b, "show db\n" );                     break;
-      case CMD_STRATS: sprintf( b, "show strat\n" );                  break;
-      case CMD_HELP:   sprintf( b, "help\n" );                        break;
+      case CMD_INFO:   snprintf( b, BUFFERSIZE, "show info %s\n",
+                                       c->database );                 break;
+      case CMD_SERVER: snprintf( b, BUFFERSIZE, "show server\n" );    break;
+      case CMD_DBS:    snprintf( b, BUFFERSIZE, "show db\n" );        break;
+      case CMD_STRATS: snprintf( b, BUFFERSIZE, "show strat\n" );     break;
+      case CMD_HELP:   snprintf( b, BUFFERSIZE, "help\n" );           break;
       case CMD_MATCH:
 	 cmd_reply.word = c->word;
-	 sprintf( b,
-		  "match %s %s \"%s\"\n",
-		  c->database, c->strategy, c->word );                break;
+	 snprintf( b, BUFFERSIZE,
+			 "match %s %s \"%s\"\n",
+			 c->database, c->strategy, c->word );         break;
       case CMD_DEFINE:
 	 cmd_reply.word = c->word;
-	 sprintf( b, "define %s \"%s\"\n", c->database, c->word );    break;
+	 snprintf( b, BUFFERSIZE, "define %s \"%s\"\n",
+			 c->database, c->word );                      break;
       case CMD_SPELL:                                                 goto end;
       case CMD_WIND:                                                  goto end;
-      case CMD_CLOSE:  sprintf( b, "quit\n" );                        break;
+      case CMD_CLOSE:  snprintf( b, BUFFERSIZE, "quit\n" );           break;
       default:
 	 err_internal( __FUNCTION__, "Unknown command %d\n", c->command );
       }
@@ -561,8 +675,9 @@ end:				/* Ready to send buffer, but are we
 	    }
 	 }
          client_close_pager();
-	 err_fatal( NULL,
-		    "Cannot connect to any servers (use -v to see why)\n" );
+	 fprintf (stderr,
+		  "Cannot connect to any servers (use -v to see why)\n");
+	 exit (EXST_CONNECTION_FAILED);
       }
       cmd_reply.host    = c->host;
       cmd_reply.service = c->service ? c->service : DICT_DEFAULT_SERVICE;
@@ -637,6 +752,7 @@ static void process( void )
 	    }
 	    xfree( cmd_reply.defs );
 	    cmd_reply.count = 0;
+
 	 } else if (cmd_reply.matches) {
 	    fprintf( dict_output,
 		     "No definitions found for \"%s\", perhaps you mean:",
@@ -646,10 +762,15 @@ static void process( void )
 	    client_free_text( cmd_reply.data );
 	    cmd_reply.data = NULL;
 	    cmd_reply.matches = 0;
+
+	    set_ex_status (EXST_APPROX_MATCHES);
 	 } else {
 	    fprintf( dict_output,
 		     "No definitions found for \"%s\"\n", c->word );
+
+	    set_ex_status (EXST_NO_MATCH);
 	 }
+
 	 expected = cmd_reply.retcode;
 	 break;
       case CMD_CONNECT:
@@ -677,6 +798,8 @@ static void process( void )
 	    fprintf( dict_output, "Client command gave unexpected status code %d (%s)\n",
 		    cmd_reply.retcode, message ? message : "no message" );
 
+//	 set_ex_status (cmd_reply.retcode);
+
 	 expected = cmd_reply.retcode;
 	 break;
       case CMD_AUTH:
@@ -762,11 +885,15 @@ static void process( void )
 		   ("No match found for \"%s\" in %s\n",c->word,c->database));
 	    break;
 	 case CODE_INVALID_DB:
-	    printf( "%s is not a valid database, use -D for a list\n",
+	    fprintf(stderr, "%s is not a valid database, use -D for a list\n",
 		    c->database );
+	    set_ex_status (EXST_INVALID_DB);
 	    break;
 	 case CODE_NO_DATABASES:
 	    fprintf( dict_output, "There are no databases currently available\n" );
+
+	    set_ex_status (EXST_NO_DATABASES);
+
 	    break;
 	 default:
 	    expected = CODE_OK;
@@ -787,28 +914,50 @@ static void process( void )
 						    &message,
 						    NULL,NULL,NULL,NULL,NULL );
 	    break;
+	 case CODE_TEMPORARILY_UNAVAILABLE:
+	    fprintf (stderr,
+		    "Server temporarily unavailable\n");
+
+	    set_ex_status (EXST_TEMPORARILY_UNAVAILABLE);
+
+	    break;
 	 case CODE_NO_MATCH:
 	    PRINTF(DBG_VERBOSE,
 		   ("No match found in %s for \"%s\" using %s\n",
 		    c->database,c->word,c->strategy));
+
+	    set_ex_status (EXST_NO_MATCH);
+
 	    break;
 	 case CODE_INVALID_DB:
 	    fprintf( dict_output,
 		     "%s is not a valid database, use -D for a list\n",
 		     c->database );
+
+	    set_ex_status (EXST_INVALID_DB);
+
 	    break;
 	 case CODE_INVALID_STRATEGY:
 	    fprintf( dict_output,
 		     "%s is not a valid search strategy, use -S for a list\n",
 		     c->strategy );
+
+	    set_ex_status (EXST_INVALID_STRATEGY);
+
 	    break;
 	 case CODE_NO_DATABASES:
 	    fprintf( dict_output,
 		     "There are no databases currently available\n" );
+
+	    set_ex_status (EXST_NO_DATABASES);
+
 	    break;
 	 case CODE_NO_STRATEGIES:
 	    fprintf( dict_output,
 		     "There are no search strategies currently available\n" );
+
+	    set_ex_status (EXST_NO_STRATEGIES);
+
 	    break;
 	 default:
 	    expected = CODE_OK;
@@ -864,6 +1013,8 @@ static void process( void )
 	 } else {
 	    fprintf( dict_output, "No matches found for \"%s\"", c->word );
 	    fprintf( dict_output, "\n" );
+
+	    set_ex_status (EXST_NO_MATCH);
 	 }
 	 expected = cmd_reply.retcode;
 	 break;
@@ -878,11 +1029,7 @@ static void process( void )
       }
       if (cmd_reply.s && cmd_reply.retcode != expected) {
          client_close_pager();
-	 err_fatal( NULL,
-		    "Unexpected status code %d (%s), wanted %d\n",
-		    cmd_reply.retcode,
-		    message ? message : "no message",
-		    expected );
+	 unexpected_status_code (expected, cmd_reply.retcode, message);
       }
       PRINTF(DBG_RAW,("* Processed %d\n",c->command));
       xfree(c);
@@ -951,7 +1098,7 @@ static const char *id_string( const char *id )
 {
    static char buffer[BUFFERSIZE];
 
-   sprintf( buffer, "%s", DICT_VERSION );
+   snprintf( buffer, BUFFERSIZE, "%s", DICT_VERSION );
    
    return buffer;
 }
@@ -959,16 +1106,15 @@ static const char *id_string( const char *id )
 static const char *client_get_banner( void )
 {
    static char       *buffer= NULL;
-   const char        *id = "$Id: dict.c,v 1.33 2003/02/02 16:34:26 hilliard Exp $";
+   const char        *id = "$Id: dict.c,v 1.43 2004/05/24 11:42:25 cheusov Exp $";
    struct utsname    uts;
    
    if (buffer) return buffer;
    uname( &uts );
    buffer = xmalloc(256);
-   sprintf( buffer,
-	    "%s %s/rf on %s %s", err_program_name(), id_string( id ),
-	    uts.sysname,
-	    uts.release );
+   snprintf( buffer, 256,
+	     "%s %s/rf on %s %s", err_program_name(), id_string( id ),
+	     uts.sysname, uts.release );
    return buffer;
 }
 
@@ -976,13 +1122,12 @@ static void banner( FILE *out_stream )
 {
    fprintf( out_stream , "%s\n", client_get_banner() );
    fprintf( out_stream,
-	    "Copyright 1997-2002 Rickard E. Faith (faith@dict.org)\n" );
+	    "Copyright 1997-2002 Rickard E. Faith (faith@dict.org)\n\n" );
 }
 
 static void license( void )
 {
    static const char *license_msg[] = {
-     "",
      "This program is free software; you can redistribute it and/or modify it",
      "under the terms of the GNU General Public License as published by the",
      "Free Software Foundation; either version 1, or (at your option) any",
@@ -1006,6 +1151,9 @@ static void license( void )
 static void help( FILE *out_stream )
 {
    static const char *help_msg[] = {
+   "Usage: dict [options] [word]",
+   "Query a dictd server for the definition of a word", 
+   "",
       "-h --host <server>        specify server",
       "-p --port <service>       specify port",
       "-d --database <dbname>    select a database to search",
@@ -1151,7 +1299,7 @@ int main( int argc, char **argv )
       char b[256];
       char *env = getenv("HOME");
 
-      sprintf( b, "%s/%s", env ? env : "./", DICT_RC_NAME );
+      snprintf( b, 256, "%s/%s", env ? env : "./", DICT_RC_NAME );
       PRINTF(DBG_VERBOSE,("Trying %s...\n",b));
       if (!access(b, R_OK)) {
 	 prs_file_nocpp(b);
@@ -1308,9 +1456,16 @@ int main( int argc, char **argv )
 					  s->user,
 					  s->secret ) );
 	 }
+#ifdef USE_DICT_ORG
       }
       append_command(make_command(CMD_CONNECT,"dict.org",     NULL,user,key));
       append_command(make_command(CMD_CONNECT,"alt0.dict.org",NULL,user,key));
+#else
+      }else{
+	 fprintf (stderr, "'dict.conf' doesn't specify any dict server\n");
+	 exit (1);
+      }
+#endif
    }
    append_command( make_command( CMD_CLIENT, client_get_banner() ) );
    if (doauth) append_command( make_command( CMD_AUTH ) );
@@ -1394,6 +1549,6 @@ int main( int argc, char **argv )
 	       tim_get_system( "total" ),
 	       client_bytes / tim_get_real( "total" ) );
    }
-   
-   return 0;
+
+   return ex_status;
 }
