@@ -19,7 +19,7 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 675 Mass Ave, Cambridge, MA 02139, USA.
  * 
- * $Id: dictP.h,v 1.6 2002/09/27 16:57:44 cheusov Exp $
+ * $Id: dictP.h,v 1.17 2004/02/24 17:55:51 cheusov Exp $
  * 
  */
 
@@ -30,29 +30,33 @@
 #include "config.h"
 #endif
 
+#include <sys/types.h>
+#include <sys/socket.h>
+#include <sys/utsname.h>
+#include <sys/wait.h>
+#include <netinet/in.h>
+#include <arpa/inet.h>
+
 #ifndef __GNUC__
 #define __FUNCTION__ __FILE__
 #endif
 
 /* AIX requires this to be the first thing in the file.  */
-#ifdef __GNUC__
-# define alloca __builtin_alloca
+#if defined(__svr4__) && defined(__sgi__) && !HAVE_ALLOCA_H /* IRIX */
+# undef HAVE_ALLOCA_H
+# define HAVE_ALLOCA_H 1
+#endif
+
+#if HAVE_ALLOCA_H
+# include <alloca.h>
 #else
-# if defined(__svr4__) && defined(__sgi__) && !HAVE_ALLOCA_H /* IRIX */
-#  undef HAVE_ALLOCA_H
-#  define HAVE_ALLOCA_H 1
-# endif
-# if HAVE_ALLOCA_H
-#  include <alloca.h>
+# ifdef _AIX
+# pragma alloca
 # else
-#  ifdef _AIX
- #pragma alloca
-#  else
-#   ifndef alloca /* predefined by HP cc +Olibcalls */
-#   if !defined(__svr4__) && !defined(__sgi__)	/* not on IRIX */
-char *alloca ();
-#   endif
-#   endif
+#  ifndef alloca /* predefined by HP cc +Olibcalls */
+#  if !defined(__svr4__) && !defined(__sgi__)	/* not on IRIX */
+    char *alloca ();
+#  endif
 #  endif
 # endif
 #endif
@@ -74,6 +78,12 @@ char *alloca ();
 # endif
 #endif
 
+#if HAVE_SIZE_T
+#include <stddef.h>
+#else
+typedef unsigned int size_t;
+#endif
+
 #if !HAVE_STRDUP
 extern char *strdup( const char * );
 #endif
@@ -86,6 +96,97 @@ extern long strtol( const char *, char **, int );
 extern unsigned long int strtoul( const char *, char **, int );
 #endif
 
+#if !HAVE_SNPRINTF
+extern int snprintf(char *str, size_t size, const char *format, ...);
+#endif
+
+#if !HAVE_VSNPRINTF
+#include <stdarg.h>
+extern int vsnprintf(char *str, size_t size, const char *format, va_list ap);
+#endif
+
+#if !HAVE_INET_ATON
+#define inet_aton(a,b) (b)->s_addr = inet_addr(a)
+#endif
+
+#if HAVE_WINT_T
+#include <wchar.h>
+#else
+typedef unsigned int wint_t;
+#endif
+
+#if HAVE_WCHAR_T
+#include <stddef.h>
+#else
+typedef unsigned int wchar_t;
+#endif
+
+#if HAVE_LANGINFO_H
+#include <langinfo.h>
+#else
+const char * nl_langinfo (int ITEM);
+#define CODESET 1234
+#endif
+
+#if HAVE_MBSTATE_T
+#include <wchar.h>
+#else
+typedef char mbstate_t;
+#endif
+
+#if !HAVE_STRLCPY
+extern size_t strlcpy (char *s, const char * wc, size_t size);
+#endif
+
+#if !HAVE_STRLCAT
+size_t strlcat(char *dst, const char *src, size_t siz);
+#endif
+
+#if !HAVE_WCRTOMB
+extern size_t wcrtomb (char *s, wchar_t wc, mbstate_t *ps);
+#endif
+
+#if !HAVE_WCTOMB
+extern int wctomb (char *s, wchar_t wc);
+#endif
+
+#if !HAVE_MBRLEN
+extern size_t mbrlen (const char *s, size_t n, mbstate_t *ps);
+#endif
+
+#if !HAVE_MBRTOWC
+extern size_t mbrtowc (wchar_t *pwc, const char *s, size_t n, mbstate_t *ps);
+#endif
+
+#if !HAVE_MBSTOWCS
+extern size_t mbstowcs (wchar_t *dest, const char *src, size_t n);
+#endif
+
+#if !HAVE_SETENV
+extern int setenv(const char *name, const char *value, int overwrite);
+#endif
+
+#if !HAVE_MBTOWC
+extern int mbtowc (wchar_t *pwc, const char *s, size_t n);
+#endif
+
+#if !HAVE_WCWIDTH
+#define wcwidth(x) (1)
+#endif
+
+#if USE_PLUGIN
+# if HAVE_DLFCN_H
+#  include <dlfcn.h>
+   typedef void *  lt_dlhandle;
+#  define lt_dlsym dlsym
+#  define lt_dlopen(filename) dlopen(filename, RTLD_NOW)
+#  define lt_dlclose dlclose
+#  define lt_dlerror dlerror
+# else
+#  include <ltdl.h>
+# endif
+#endif
+
 /* Get time functions */
 #if TIME_WITH_SYS_TIME
 # include <sys/time.h>
@@ -156,6 +257,19 @@ situations that we know about. */
 
 #if HAVE_LIMITS_H
 #include <limits.h>
+#endif
+
+/* Handle getopt correctly */
+#if HAVE_GETOPT_H
+# include <getopt.h>
+#else
+#if !HAVE_GETOPT_LONG
+int getopt_long(int argc, char * const argv[],
+                  const char *optstring,
+                  const struct option *longopts, int *longindex);
+extern int  optind;
+extern char *optarg;
+#endif
 #endif
 
 				/* Local stuff */
@@ -163,4 +277,12 @@ situations that we know about. */
 #define max(a,b) ((a)>(b)?(a):(b))
 #endif
 
+#if HAVE_ISWALNUM && HAVE_ISWSPACE && HAVE_TOWLOWER
+#ifdef HAVE_UTF8
+#undef HAVE_UTF8
 #endif
+
+#define HAVE_UTF8 1
+#endif
+
+#endif /* _DICTP_H_ */
