@@ -17,13 +17,22 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 675 Mass Ave, Cambridge, MA 02139, USA.
  * 
- * $Id: dictd.c,v 1.79 2003/03/09 17:07:36 cheusov Exp $
+ * $Id: dictd.c,v 1.113 2004/05/30 15:55:05 cheusov Exp $
  * 
  */
 
 #include "dictd.h"
+#include "str.h"
+
 #include "servparse.h"
 #include "strategy.h"
+#include "index.h"
+#include "data.h"
+#include "parse.h"
+
+#ifdef USE_PLUGIN
+#include "plugin.h"
+#endif
 
 #include <grp.h>                /* initgroups */
 #include <pwd.h>                /* getpwuid */
@@ -40,22 +49,8 @@
 #define GID_NOGROUP 65534
 #endif
 
-#ifndef HAVE_SNPRINTF
-extern int snprintf(char *str, size_t size, const char *format, ...);
-#endif
-
-#ifndef HAVE_VSNPRINTF
-extern int vsnprintf(char *str, size_t size, const char *format, va_list ap);
-#endif
-
 extern int        yy_flex_debug;
 
-extern int        default_strategy;
-
-extern int        utf8_mode;
-extern int        bit8_mode;
-extern int        mmap_mode;
-
 static int        _dict_daemon;
 static int        _dict_reaps;
 static int        _dict_daemon_limit        = DICT_DAEMON_LIMIT;
@@ -64,9 +59,24 @@ static char       *_dict_argvstart;
 static int        _dict_argvlen;
 
        int        _dict_forks;
+const char        *locale       = "C";
+const char        *preprocessor = NULL;
+const char        *bind_to      = NULL;
+int                inetd        = 0;
+
+int                need_reload_config    = 0;
+
+int                need_unload_databases = 0;
+int                databases_unloaded    = 0;
 
 static const char *configFile  = DICT_CONFIG_PATH DICTD_CONFIG_NAME;
 
+static void dict_close_databases (dictConfig *c);
+static void sanity (const char *confFile);
+static void dict_init_databases (dictConfig *c);
+static void dict_config_print (FILE *stream, dictConfig *c);
+static void postprocess_filenames (dictConfig *dc);
+
 void dict_initsetproctitle( int argc, char **argv, char **envp )
 {
    int i;
@@ -133,12 +143,23 @@ const char *dict_format_time( double t )
    return this;
 }
 
+static int waitpid__exit_status (int status)
+{
+   if (WIFEXITED(status)){
+      return WEXITSTATUS(status);
+   }else if (WIFSIGNALED(status)){
+      return 128 + WTERMSIG(status);
+   }else{
+      return -1;
+   }
+}
+
 static void reaper( int dummy )
 {
-#if defined(__osf__) || (defined(__sparc) && defined(__SVR4))
-   int        status;
-#else
+#if 0
    union wait status;
+#else
+   int        status;
 #endif
    pid_t      pid;
 
@@ -146,9 +167,10 @@ static void reaper( int dummy )
       ++_dict_reaps;
       
       if (flg_test(LOG_SERVER))
-         log_info( ":I: Reaped %d%s\n",
+         log_info( ":I: Reaped %d%s, exit status %i\n",
                    pid,
-                   _dict_daemon ? " IN CHILD": "" );
+                   _dict_daemon ? " IN CHILD": "",
+		   waitpid__exit_status (status));
    }
 }
 
@@ -212,6 +234,29 @@ static void log_sig_info (int sig)
       dict_format_time (tim_get_system ("dictd")));
 }
 
+static void unload_databases (void)
+{
+   dict_close_databases (DictConfig);
+   DictConfig = NULL;
+}
+
+static void reload_config (void)
+{
+   dict_close_databases (DictConfig);
+
+   if (!access(configFile,R_OK)){
+      prs_file_pp (preprocessor, configFile);
+      postprocess_filenames (DictConfig);
+   }
+
+   sanity (configFile);
+
+   if (dbg_test (DBG_VERBOSE))
+      dict_config_print( NULL, DictConfig );
+
+   dict_init_databases (DictConfig);
+}
+
 static void handler( int sig )
 {
    const char *name = NULL;
@@ -223,17 +268,22 @@ static void handler( int sig )
       daemon_terminate( sig, name );
    } else {
       tim_stop( "dictd" );
-      if (sig == SIGALRM && _dict_markTime > 0) {
-	 time(&t);
-	 log_info( ":T: %24.24s; %d/%d %sr %su %ss\n",
-		   ctime(&t),
-		   _dict_forks - _dict_reaps,
-		   _dict_forks,
-		   dict_format_time( tim_get_real( "dictd" ) ),
-		   dict_format_time( tim_get_user( "dictd" ) ),
-		   dict_format_time( tim_get_system( "dictd" ) ) );
-	 alarm(_dict_markTime);
-	 return;
+      switch (sig){
+      case SIGALRM:
+	 if (_dict_markTime > 0){
+	    time(&t);
+	    log_info( ":T: %24.24s; %d/%d %sr %su %ss\n",
+		      ctime(&t),
+		      _dict_forks - _dict_reaps,
+		      _dict_forks,
+		      dict_format_time( tim_get_real( "dictd" ) ),
+		      dict_format_time( tim_get_user( "dictd" ) ),
+		      dict_format_time( tim_get_system( "dictd" ) ) );
+	    alarm(_dict_markTime);
+	    return;
+	 }
+
+	 break;
       }
 
       log_sig_info (sig);
@@ -242,11 +292,6 @@ static void handler( int sig )
       exit(sig+128);
 }
 
-static void dict_close_databases (dictConfig *c);
-static void sanity (const char *confFile);
-static void dict_init_databases (dictConfig *c);
-static void dict_config_print (FILE *stream, dictConfig *c);
-
 static const char *postprocess_filename (const char *fn, const char *prefix)
 {
    char *new_fn;
@@ -294,20 +339,13 @@ static void postprocess_filenames (dictConfig *dc)
 static void handler_sighup (int sig)
 {
    log_sig_info (sig);
+   need_reload_config = 1;
+}
 
-   dict_close_databases (DictConfig);
-
-   if (!access(configFile,R_OK)){
-      prs_file_nocpp (configFile);
-      postprocess_filenames (DictConfig);
-   }
-
-   sanity (configFile);
-
-   if (dbg_test (DBG_VERBOSE))
-      dict_config_print( NULL, DictConfig );
-
-   dict_init_databases (DictConfig);
+static void handler_sigusr1 (int sig)
+{
+   log_sig_info (sig);
+   need_unload_databases = 1;
 }
 
 static void setsig( int sig, void (*f)(int), int sa_flags )
@@ -391,7 +429,7 @@ static int config_print( const void *datum, void *arg )
    if (db->indexwordFilename)
       fprintf( s, "   index_word      %s\n", db->indexwordFilename );
    if (db->filter)
-      fprintf( s, "   filter     %s\n", db->filter ); /* Not implemented. */
+      fprintf( s, "   filter     %s\n", db->filter );
    if (db->prefilter)
       fprintf( s, "   prefilter  %s\n", db->prefilter );
    if (db->postfilter)
@@ -426,6 +464,9 @@ static const char *get_entry_info( dictDatabase *db, const char *entryName )
 	 list, entryName, db, DICT_EXACT,
 	 NULL, NULL, NULL ))
    {
+#ifdef USE_PLUGIN
+      call_dictdb_free (DictConfig->dbl);
+#endif
       lst_destroy( list );
       return NULL;
    }
@@ -434,16 +475,21 @@ static const char *get_entry_info( dictDatabase *db, const char *entryName )
 
    buf = pt = dict_data_obtain( db, dw );
 
-   while (*pt != '\n')
-      ++pt;
+   if (!strncmp (pt, "00database", 10) || !strncmp (pt, "00-database", 11)){
+      while (*pt != '\n')
+	 ++pt;
 
-   ++pt;
+      ++pt;
+   }
 
    while (*pt == ' ' || *pt == '\t')
       ++pt;
 
    pt[ strlen(pt) - 1 ] = '\0';
 
+#ifdef USE_PLUGIN
+   call_dictdb_free (DictConfig->dbl);
+#endif
    dict_destroy_list( list );
 
    pt = xstrdup (pt);
@@ -549,22 +595,77 @@ static int init_plugin( const void *datum )
 {
 #ifdef USE_PLUGIN
    dictDatabase *db = (dictDatabase *)datum;
-   dict_plugin_open (db);
+   dict_plugin_init (db);
 #endif
 
    return 0;
 }
 
+void dict_disable_strat (dictDatabase *db, const char* strategy)
+{
+   int strat = -1;
+   int array_size = get_max_strategy_num () + 1;
+
+   assert (db);
+   assert (strategy);
+
+   if (!db -> strategy_disabled){
+      db -> strategy_disabled = xmalloc (array_size * sizeof (int));
+      memset (db -> strategy_disabled, 0, array_size * sizeof (int));
+   }
+
+   strat = lookup_strategy (strategy);
+
+   if (strat == -1){
+      log_info(":E: strategy '%s' is not available\n", strategy);
+      err_fatal(__FUNCTION__, ":E: terminating due to errors.\n");
+   }else{
+      db -> strategy_disabled [strat] = 1;
+   }
+}
+
+static void init_database_alphabet (dictDatabase *db)
+{
+   int ret;
+   lst_List l;
+   const dictWord *dw;
+   char *data;
+
+   if (!db -> normal_db)
+      return;
+
+   l = lst_create ();
+
+   ret = dict_search_database_ (l, DICT_FLAG_ALPHABET, db, DICT_EXACT);
+
+   if (ret){
+      dw = (const dictWord *) lst_top (l);
+      data = dict_data_obtain (db, dw);
+      db -> alphabet = data;
+
+      data = strchr (db -> alphabet, '\n');
+      if (data)
+	 *data = 0;
+   }
+
+   dict_destroy_list (l);
+}
+
 static int init_database( const void *datum )
 {
    dictDatabase *db = (dictDatabase *)datum;
 
    PRINTF (DBG_INIT, (":I: Initializing '%s'\n", db->databaseName));
 
-   PRINTF (DBG_INIT, (":I:   Opening indices\n"));
+   if (db->indexFilename){
+      PRINTF (DBG_INIT, (":I:   Opening indices\n"));
+   }
 
    db->index        = dict_index_open( db->indexFilename, 1, 0, 0 );
-   PRINTF (DBG_INIT, (":I:     .index <ok>\n"));
+
+   if (db->indexFilename){
+      PRINTF (DBG_INIT, (":I:     .index <ok>\n"));
+   }
 
    if (db->index){
       db->index_suffix = dict_index_open(
@@ -588,25 +689,46 @@ static int init_database( const void *datum )
       db->index_word->flag_allchars = db->index->flag_allchars;
    }
 
-   PRINTF (DBG_INIT, (":I:   Opening data\n"));
+   if (db->dataFilename){
+      PRINTF (DBG_INIT, (":I:   Opening data\n"));
+   }
+
    db->data         = dict_data_open( db->dataFilename, 0 );
 
-   PRINTF(DBG_INIT,
-	  (":I: '%s' initialized\n", db->databaseName));
+   init_database_alphabet (db);
+   if (db -> alphabet){
+      PRINTF (DBG_INIT, (":I:     alphabet: %s\n", db -> alphabet));
+   }else{
+      PRINTF (DBG_INIT, (":I:     alphabet: (NULL)\n"));
+   }
+
+   if (db->dataFilename){
+      PRINTF(DBG_INIT,
+	     (":I: '%s' initialized\n", db->databaseName));
+   }
 
    return 0;
 }
 
 static int init_database_short (const void *datum)
 {
-   dictDatabase *db = (dictDatabase *)datum;
+   char *NL;
 
-   if (!db->databaseShort)
+   dictDatabase *db = (dictDatabase *) datum;
+
+   if (!db->databaseShort){
       db->databaseShort = get_entry_info( db, DICT_SHORT_ENTRY_NAME );
-   else if (*db->databaseShort == '@' && !db -> virtual_db)
+   }else if (*db->databaseShort == '@'){
       db->databaseShort = get_entry_info( db, db->databaseShort + 1 );
-   else
+   }else{
       db->databaseShort = xstrdup (db->databaseShort);
+   }
+
+   if (db->databaseShort){
+      NL = strchr (db->databaseShort, '\n');
+      if (NL)
+	 *NL = 0;
+   }
 
    if (!db->databaseShort)
       db->databaseShort = xstrdup (db->databaseName);
@@ -618,7 +740,7 @@ static int close_plugin (const void *datum)
 {
 #ifdef USE_PLUGIN
    dictDatabase  *db = (dictDatabase *)datum;
-   dict_plugin_close (db);
+   dict_plugin_destroy (db);
 #endif
 
    return 0;
@@ -647,6 +769,10 @@ static int close_database (const void *datum)
       xfree ((void *) db -> indexsuffixFilename);
    if (db -> pluginFilename)
       xfree ((void *) db -> pluginFilename);
+   if (db -> strategy_disabled)
+      xfree ((void *) db -> strategy_disabled);
+   if (db -> alphabet)
+      xfree ((void *) db -> alphabet);
 
    return 0;
 }
@@ -672,7 +798,7 @@ static int log_database_info( const void *datum )
 
 static void dict_ltdl_init ()
 {
-#ifdef USE_PLUGIN
+#if USE_PLUGIN && !HAVE_DLFCN_H
    if (lt_dlinit ())
       err_fatal( __FUNCTION__, "Can not initialize 'ltdl' library\n" );
 #endif
@@ -680,7 +806,7 @@ static void dict_ltdl_init ()
 
 static void dict_ltdl_close ()
 {
-#ifdef USE_PLUGIN
+#if USE_PLUGIN && !HAVE_DLFCN_H
    if (lt_dlexit ())
       err_fatal( __FUNCTION__, "Can not deinitialize 'ltdl' library\n" );
 #endif
@@ -726,6 +852,9 @@ static void dict_close_databases (dictConfig *c)
    dictDatabase *db;
    dictAccess   *acl;
 
+   if (!c)
+      return;
+
    if (c -> dbl){
       while (lst_length (c -> dbl) > 0){
 	 db = (dictDatabase *) lst_pop (c -> dbl);
@@ -780,21 +909,9 @@ static int dump_def( const void *datum )
    return 0;
 }
 
-static int call_dictdb_free (const void *datum)
-{
-   const dictWord     *dw = (dictWord *)datum;
-   const dictDatabase *db = dw -> database;
-
-   if (db -> plugin)
-      db -> plugin -> dictdb_free (db -> plugin -> data);
-
-   return 0;
-}
-
 static void dict_dump_defs( lst_List list )
 {
    lst_iterate (list, dump_def);
-   lst_iterate (list, call_dictdb_free);
 }
 
 static const char *id_string( const char *id )
@@ -810,7 +927,7 @@ const char *dict_get_banner( int shortFlag )
 {
    static char    *shortBuffer = NULL;
    static char    *longBuffer = NULL;
-   const char     *id = "$Id: dictd.c,v 1.79 2003/03/09 17:07:36 cheusov Exp $";
+   const char     *id = "$Id: dictd.c,v 1.113 2004/05/30 15:55:05 cheusov Exp $";
    struct utsname uts;
    
    if (shortFlag && shortBuffer) return shortBuffer;
@@ -839,13 +956,12 @@ const char *dict_get_banner( int shortFlag )
 static void banner( void )
 {
    printf( "%s\n", dict_get_banner(0) );
-   printf( "Copyright 1997-2002 Rickard E. Faith (faith@dict.org)\n" );
+   printf( "Copyright 1997-2002 Rickard E. Faith (faith@dict.org)\n\n" );
 }
 
 static void license( void )
 {
    static const char *license_msg[] = {
-     "",
      "This program is free software; you can redistribute it and/or modify it",
      "under the terms of the GNU General Public License as published by the",
      "Free Software Foundation; either version 1, or (at your option) any",
@@ -869,6 +985,9 @@ static void license( void )
 static void help( void )
 {
    static const char *help_msg[] = {
+   "Usage: dictd [options]",
+   "Start the dictd daemon",
+   "",
       "-h --help             give this help",
       "   --license          display software license",
       "-v --verbose          verbose mode",
@@ -884,6 +1003,8 @@ static void help( void )
       "-m --mark <minutes>   how often should a timestamp be logged",
       "   --facility <fac>   set syslog logging facility",
       "-d --debug <option>   select debug option",
+      "-i --inetd            run from inetd",
+      "   --pp <prog>        set preprocessor for configuration file",
       "-f --force            force startup even if daemon running",
       "   --locale <locale>  specifies the locale used for searching.\n\
                       if no locale is specified, the \"C\" locale is used.",
@@ -893,10 +1014,7 @@ static void help( void )
                                    <strategies> is a comma-separated list.",
 "   --add-strategy <strat>:<descr>  adds new strategy <strat>\n\
                                    with a description <descr>.",
-#ifdef HAVE_MMAP
-"   --no-mmap          do not use mmap() function and load files\n\
-                      into memory instead.",
-#endif
+"   --listen-to                     bind a socket to the specified address",
 "\n------------------ options for debugging ---------------------------",
 "-t --test <word>                lookup word",
 "   --test-file <file>",
@@ -908,6 +1026,11 @@ static void help( void )
 "   --test-match                 show matched words but the definitions",
 "   --test-nooutput              produces no output",
 "   --test-idle                  does everything except search",
+"   --fast-start                 don't create additional index.",
+#ifdef HAVE_MMAP
+"   --without-mmap               do not use mmap() function and load files\n\
+                                into memory instead.",
+#endif
       0 };
    const char        **p = help_msg;
 
@@ -1048,33 +1171,38 @@ static void sanity(const char *confFile)
 		  getegid(), gr && gr->gr_name ? gr->gr_name : "?");
 	 log_info(":E: config and db files must be readable by that user\n");
       }
-      err_fatal(__FUNCTION__, ":E: terminating due to errors\n");
+      err_fatal(__FUNCTION__, ":E: terminating due to errors. See log file\n");
    }
 }
 
-static char *strlwr_8bit (char *str)
+static void set_locale_and_flags (const char *loc)
 {
-   char *p;
-   for (p = str; *p; ++p){
-      *p = tolower ((unsigned char) *p);
+   const char *charset = NULL;
+   int ascii_mode;
+
+   if (!setlocale(LC_COLLATE, loc) || !setlocale(LC_CTYPE, loc)){
+      fprintf (stderr, "invalid locale '%s'\n", locale);
+      exit (2);
    }
 
-   return str;
-}
+   charset = nl_langinfo (CODESET);
 
-static void set_utf8bit_mode (const char *loc)
-{
-   char *locale_copy;
-   locale_copy = strdup (loc);
-   strlwr_8bit (locale_copy);
+   utf8_mode = !strcmp (charset, "UTF-8");
 
-   utf8_mode =
-       strstr (locale_copy, "utf-8") ||
-       strstr (locale_copy, "utf8");
+#if !HAVE_UTF8
+   if (utf8_mode){
+      err_fatal (
+	 __FUNCTION__,
+	 "utf-8 support was disabled at compile time\n");
+   }
+#endif
 
-   bit8_mode = !utf8_mode && (locale_copy [0] != 'c' || locale_copy [1] != 0);
+   ascii_mode = 
+      !strcmp (charset, "ANSI_X3.4-1968") ||
+      !strcmp (charset, "US-ASCII") ||
+      (locale [0] == 'C' && locale [1] == 0);
 
-   free (locale_copy);
+   bit8_mode = !ascii_mode && !utf8_mode;
 }
 
 static void init (const char *fn)
@@ -1086,12 +1214,8 @@ static void init (const char *fn)
 
 static void destroy ()
 {
-   /*
-     tim_stop ("dictd");
-     maa_shutdown ();
-   */
-   src_destroy ();
-   str_destroy ();
+   maa_shutdown ();
+
    dict_ltdl_close ();
    dict_destroy_strategies ();
 }
@@ -1123,19 +1247,108 @@ static void dict_test (
 
    count = dict_search_databases (l, NULL, database_arg, word, strategy, &db_found);
 
-   if (!nooutput_mode){
-      if (count != 0){
-	 dict_dump_defs (l);
-      }else{
-	 fprintf (stderr, "No definitions found for \"%s\"\n", word);
+   if (db_found){
+      if (!nooutput_mode){
+	 if (count > 0){
+	    dict_dump_defs (l);
+	 }else{
+	    fprintf (stderr, "No definitions found for \"%s\"\n", word);
+	 }
       }
+   }else{
+      fprintf (stderr, "%s is not a valid database\n", database_arg);
    }
 
+#ifdef USE_PLUGIN
+   call_dictdb_free (DictConfig->dbl);
+#endif
+
    dict_destroy_list (l);
 }
 
-//static const char *       without_strategy_arg = NULL;
-//static dictStrategy *     without_strategy = (dictStrategy *) 1;
+static void dict_test_word (const char *word, int strategy)
+{
+   dict_config_print( NULL, DictConfig );
+   dict_init_databases( DictConfig );
+
+   dict_make_dbs_available (DictConfig);
+
+   if (!idle_mode){
+      dict_test (word, strategy);
+
+      if (!nooutput_mode){
+	 fprintf( stderr, "%d comparisons\n", _dict_comparisons );
+      }
+   }
+
+   dict_close_databases (DictConfig);
+
+   destroy ();
+
+   exit( 0 );
+}
+
+static void dict_test_file (const char *filename, int strategy)
+{
+   FILE         *str;
+   char         buf[1024], *pt;
+   int          words = 0;
+   int                word_len;
+
+   if (!(str = fopen(filename,"r")))
+      err_fatal_errno( "Cannot open \"%s\" for read\n", filename );
+
+   dict_config_print( NULL, DictConfig );
+   dict_init_databases( DictConfig );
+   dict_make_dbs_available (DictConfig);
+
+   while (fgets(buf,1024,str)) {
+      word_len = strlen( buf );
+      if (word_len > 0){
+	 if ('\n' == buf [word_len - 1]){
+	    buf [word_len - 1] = '\0';
+	 }
+      }
+
+      if ((pt = strchr(buf, '\t')))
+	 *pt = '\0'; /* stop at tab */
+
+      if (buf[0]){
+	 ++words;
+
+	 if (!idle_mode){
+	    dict_test (buf, strategy);
+	 }
+      }
+
+      if (words && !(words % 1000)){
+	 if (!nooutput_mode){
+	    fprintf(
+	       stderr,
+	       "%d comparisons, %d words\n", _dict_comparisons, words );
+	 }
+      }
+   }
+
+   if (!nooutput_mode){
+      fprintf(
+	 stderr,
+	 "%d comparisons, %d words\n", _dict_comparisons, words );
+   }
+
+   fclose( str);
+
+   dict_close_databases (DictConfig);
+
+   destroy ();
+
+   exit(0);
+/* Comparisons:
+   P5/133
+   1878064 comparisons, 113955 words
+   39:18.72u 1.480s 55:20.27 71%
+*/
+}
 
 int main( int argc, char **argv, char **envp )
 {
@@ -1144,7 +1357,6 @@ int main( int argc, char **argv, char **envp )
    struct sockaddr_in csin;
    int                c;
    time_t             startTime;
-   int                word_len;
    int                alen         = sizeof(csin);
    const char         *service     = DICT_DEFAULT_SERVICE;
    int                detach       = 1;
@@ -1156,7 +1368,6 @@ int main( int argc, char **argv, char **envp )
    int                useSyslog    = 0;
    int                logOptions   = 0;
    int                forceStartup = 0;
-   const char         *locale      = "C";
    int                i;
 
    const char *       strategy_arg = "exact";
@@ -1187,10 +1398,12 @@ int main( int argc, char **argv, char **envp )
       { "limit",    1, 0, 504 },
       { "facility", 1, 0, 505 },
       { "force",    1, 0, 'f' },
+      { "inetd",    0, 0, 'i' },
       { "locale",           1, 0, 506 },
       { "test-strategy",    1, 0, 507 },
 #ifdef HAVE_MMAP
       { "no-mmap",          0, 0, 508 },
+      { "without-mmap",     0, 0, 508 },
 #endif
       { "test-db",          1, 0, 509 },
       { "default-strategy", 1, 0, 511 },
@@ -1199,12 +1412,13 @@ int main( int argc, char **argv, char **envp )
       { "test-nooutput",    0, 0, 514 },
       { "test-idle",        0, 0, 515 },
       { "add-strategy",     1, 0, 516 },
+      { "fast-start",       0, 0, 517 },
+      { "pp",               1, 0, 518 },
+      { "listen-to",        1, 0, 519 },
       { 0,                  0, 0, 0  }
    };
 
-   release_root_privileges();
-
-   init(argv[0]);
+   init (argv[0]);
 
    flg_register( LOG_SERVER,    "server" );
    flg_register( LOG_CONNECT,   "connect" );
@@ -1232,7 +1446,7 @@ int main( int argc, char **argv, char **envp )
    dbg_register( DBG_ALT,      "alt" );
 
    while ((c = getopt_long( argc, argv,
-			    "vVd:p:c:hL:t:l:sm:f", longopts, NULL )) != EOF)
+			    "vVd:p:c:hL:t:l:sm:fi", longopts, NULL )) != EOF)
       switch (c) {
                                 /* Remember to copy optarg since we're
                                    going to destroy argv soon... */
@@ -1246,6 +1460,10 @@ int main( int argc, char **argv, char **envp )
       case 's': ++useSyslog;                              break;
       case 'm': _dict_markTime = 60*atoi(optarg);         break;
       case 'f': ++forceStartup;                           break;
+      case 'i':
+	 inetd         = 1;
+	 optStart_mode = 0;
+	 break;
       case 'l':
 	 ++logOptions;
 	 flg_set( optarg );
@@ -1257,16 +1475,16 @@ int main( int argc, char **argv, char **envp )
       case 503: depth = atoi(optarg);                     break;
       case 504: _dict_daemon_limit = atoi(optarg);        break;
       case 505: ++useSyslog; log_set_facility(optarg);    break;
-      case 506: locale = optarg;                          break;
+      case 506: locale = str_copy(optarg);                break;
       case 508: mmap_mode = 0;                            break;
-      case 509: database_arg = optarg;                    break;
+      case 509: database_arg = str_copy(optarg);          break;
       case 507:
-	 strategy_arg = optarg;
-	 strategy = lookup_strategy(optarg);
+	 strategy_arg = str_copy (optarg);
+	 strategy = lookup_strategy (strategy_arg);
 	 break;
       case 511:
-	 default_strategy_arg = optarg;
-	 default_strategy = lookup_strategy(optarg);
+	 default_strategy_arg = str_copy (optarg);
+	 default_strategy = lookup_strategy (default_strategy_arg);
 	 break;
       case 512:
 	 match_mode = 1;
@@ -1293,13 +1511,18 @@ int main( int argc, char **argv, char **envp )
 	 dict_add_strategy (new_strategy, new_strategy_descr);
 
 	 break;
+      case 517: optStart_mode = 0;                        break;
+      case 518: preprocessor = str_copy (optarg);         break;
+      case 519: bind_to      = str_copy (optarg);         break;
       case 'h':
       default:  help(); exit(0);                          break;
       }
 
+   if (testWord || testFile || inetd)
+      detach = 0;
+
    if (
       -1 == strategy ||
-//      (strategy_arg = without_strategy_arg, NULL == without_strategy) ||
       (strategy_arg = default_strategy_arg, -1 == default_strategy))
    {
       fprintf (stderr, "%s is not a valid search strategy\n", strategy_arg);
@@ -1320,120 +1543,46 @@ int main( int argc, char **argv, char **envp )
    if (flg_test(LOG_TIMESTAMP)) log_option( LOG_OPTION_FULL );
    else                         log_option( LOG_OPTION_NO_FULL );
 
-   set_utf8bit_mode (locale);
+   if (! inetd && detach)
+      net_detach();
 
-   if (!setlocale(LC_ALL, locale)){
-      fprintf (stderr, "invalid locale '%s'\n", locale);
-      exit (2);
-   }
+   if (logFile) log_file( "dictd", logFile );
+   if (useSyslog) log_syslog( "dictd" );
+
+   release_root_privileges();
+
+   set_locale_and_flags (locale);
 
    time(&startTime);
    tim_start( "dictd" );
    alarm(_dict_markTime);
 
    if (!access(configFile,R_OK)) {
-      prs_file_nocpp( configFile );
+      prs_file_pp (preprocessor, configFile );
       postprocess_filenames (DictConfig);
    }
 
-
-                                /* Open up logs for sanity check */
-   if (logFile)   log_file( "dictd", logFile );
-   if (useSyslog) log_syslog( "dictd" );
-   log_stream( "dictd", stderr );
+/*   log_stream( "dictd", stderr );*/
    sanity(configFile);
-   log_close();
-
+/*   log_stream( "dictd", NULL );*/
 
    if (match_mode)
       strategy |= DICT_MATCH_MASK;
 
 
    if (testWord) {		/* stand-alone test mode */
-      dict_config_print( NULL, DictConfig );
-      dict_init_databases( DictConfig );
-
-      dict_make_dbs_available (DictConfig);
-
-      if (!idle_mode){
-	 dict_test (testWord, strategy);
-
-	 if (!nooutput_mode){
-	    fprintf( stderr, "%d comparisons\n", _dict_comparisons );
-	 }
-      }
-
-      dict_close_databases (DictConfig);
-
-      destroy ();
-
-      exit( 0 );
+      dict_test_word (testWord, strategy);
+      abort (); /* this should not happen */
    }
 
    if (testFile) {
-      FILE         *str;
-      char         buf[1024], *pt;
-      int          words = 0;
-
-      if (!(str = fopen(testFile,"r")))
-	 err_fatal_errno( "Cannot open \"%s\" for read\n", testFile );
-
-      dict_config_print( NULL, DictConfig );
-      dict_init_databases( DictConfig );
-      dict_make_dbs_available (DictConfig);
-
-      while (fgets(buf,1024,str)) {
-         word_len = strlen( buf );
-         if (word_len > 0){
-            if ('\n' == buf [word_len - 1]){
-               buf [word_len - 1] = '\0';
-            }
-         }
-
-         if ((pt = strchr(buf, '\t')))
-	    *pt = '\0'; /* stop at tab */
-
-         if (buf[0]){
-	    ++words;
-
-	    if (!idle_mode){
-	       dict_test (buf, strategy);
-	    }
-         }
-
-         if (words && !(words % 1000)){
-	    if (!nooutput_mode){
-	       fprintf(
-		  stderr,
-		  "%d comparisons, %d words\n", _dict_comparisons, words );
-	    }
-	 }
-      }
-
-      if (!nooutput_mode){
-	 fprintf(
-	    stderr,
-	    "%d comparisons, %d words\n", _dict_comparisons, words );
-      }
-
-      fclose( str);
-
-      dict_close_databases (DictConfig);
-
-      destroy ();
-
-      exit(0);
-      /* Comparisons:
-	 P5/133
-	 1878064 comparisons, 113955 words
-	 39:18.72u 1.480s 55:20.27 71%
-	 */
-
-	
+      dict_test_file (testFile, strategy);
+      abort (); /* this should not happen */
    }
 
    setsig(SIGCHLD, reaper, SA_RESTART);
-   setsig(SIGHUP,  handler_sighup, 0);
+   setsig(SIGHUP,   handler_sighup, 0);
+   setsig(SIGUSR1,  handler_sigusr1, 0);
    if (!dbg_test(DBG_NOFORK))
       setsig(SIGINT,  handler, 0);
    setsig(SIGQUIT, handler, 0);
@@ -1446,25 +1595,30 @@ int main( int argc, char **argv, char **envp )
    fflush(stdout);
    fflush(stderr);
 
-   if (detach) net_detach();
-
-                                /* Re-open logs for logging */
-   if (logFile)   log_file( "dictd", logFile );
-   if (useSyslog) log_syslog( "dictd" );
    if (!detach)   log_stream( "dictd", stderr );
-   if ((logFile || useSyslog || !detach) && !logOptions) set_minimal();
+   if ((logFile || useSyslog || !detach) && !logOptions)
+      set_minimal();
 
    log_info(":I: %d starting %s %24.24s\n",
 	    getpid(), dict_get_banner(0), ctime(&startTime));
-   if (strcmp(locale, "C")) log_info(":I: using locale \"%s\"\n", locale);
+   if (strcmp(locale, "C"))
+      log_info(":I: using locale \"%s\"\n", locale);
+
+   if (dbg_test(DBG_VERBOSE))
+      dict_config_print( NULL, DictConfig );
 
-   masterSocket = net_open_tcp( service, depth );
-   
-   if (dbg_test(DBG_VERBOSE)) dict_config_print( NULL, DictConfig );
    dict_init_databases( DictConfig );
 
    dict_initsetproctitle(argc, argv, envp);
 
+   if (inetd) {
+      dict_inetd(&argv,delay,0);
+      exit(0);
+   }
+
+   masterSocket = net_open_tcp( bind_to, service, depth );
+
+
    for (;;) {
       dict_setproctitle( "%s: %d/%d",
 			 dict_get_banner(1),
@@ -1474,7 +1628,21 @@ int main( int argc, char **argv, char **envp )
          log_info( ":I: %d accepting on %s\n", getpid(), service );
       if ((childSocket = accept(masterSocket,
 				(struct sockaddr *)&csin, &alen)) < 0) {
-	 if (errno == EINTR) continue;
+	 if (errno == EINTR){
+	    if (need_reload_config){
+	       reload_config ();
+	       need_reload_config = 0;
+	       databases_unloaded = 0;
+	    }
+
+	    if (need_unload_databases){
+	       unload_databases ();
+	       need_unload_databases = 0;
+	       databases_unloaded = 1;
+	    }
+	    continue;
+	 }
+
 #ifdef __linux__
 				/* Linux seems to return more types of
                                    errors than other OSs. */
@@ -1494,8 +1662,12 @@ int main( int argc, char **argv, char **envp )
       } else {
 	 if (_dict_forks - _dict_reaps < _dict_daemon_limit) {
 	    if (!start_daemon()) { /* child */
+	       int databases_loaded = (DictConfig != NULL);
+
 	       alarm(0);
-	       dict_daemon(childSocket,&csin,&argv,delay,0);
+	       dict_daemon (
+		  childSocket, &csin, &argv, delay,
+		  databases_loaded ? 0 : 2);
 	       exit(0);
 	    } else {		   /* parent */
 	       close(childSocket);
