@@ -17,36 +17,46 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 675 Mass Ave, Cambridge, MA 02139, USA.
  * 
- * $Id: index.c,v 1.61 2003/03/09 17:06:16 cheusov Exp $
+ * $Id: index.c,v 1.89 2004/03/22 17:17:06 cheusov Exp $
  * 
  */
 
+#include "dictP.h"
 #include "dictzip.h"
-#include "dictd.h"
-#include "regex.h"
+#include "index.h"
+#include "include_regex.h"
 #include "strategy.h"
+#include "str.h"
+
+#ifdef USE_PLUGIN
+#include "plugin.h"
+#endif
 
 #include <sys/stat.h>
 
 #include <fcntl.h>
 #include <ctype.h>
-#include <wctype.h>
-#include <wchar.h>
-#include <stdio.h>
 
-#ifdef USE_PLUGIN
-#include <ltdl.h>
+#if HAVE_WCTYPE_H
+#include <wctype.h>
 #endif
 
+#if HAVE_WCHAR_H
+#include <wchar.h>
+#endif
+
+#include <stdio.h>
+
 extern int mmap_mode;
 
 #define FIND_NEXT(pt,end) while (pt < end && *pt++ != '\n');
-#define OPTSTART        1	/* Optimize search range for constant start */
 #define MAXWORDLEN    512
 #define BMH_THRESHOLD   3	/* When to start using Boyer-Moore-Hoorspool */
 
-int utf8_mode;     /* dictd uses UTF-8 dictionaries */
-int bit8_mode;     /* dictd uses 8-BIT dictionaries */
+int utf8_mode;          /* dictd uses UTF-8 dictionaries */
+int bit8_mode;          /* dictd uses 8-BIT dictionaries */
+int optStart_mode = 1;	/* Optimize search range for constant start */
+
 dictConfig *DictConfig;
 
 int _dict_comparisons;
@@ -56,6 +66,9 @@ static int isspacepuncttab [UCHAR_MAX + 1];
 static int char2indextab[UCHAR_MAX + 2];
 static int index2chartab[UCHAR_MAX + 2];
 
+char global_alphabet_8bit [UCHAR_MAX + 2];
+char global_alphabet_ascii [UCHAR_MAX + 2];
+
 static int chartab [UCHAR_MAX + 1];
 static int charcount = 0;
 
@@ -91,69 +104,28 @@ static int dict_table_init_compare_utf8 (const void *a, const void *b)
     return c1 - c2;
 }
 
-/*
-  Copies alphanumeric and space characters converting them to lower case.
-  Strings are represented in 8-bit character set.
-*/
-static int tolower_alnumspace_8bit (
-   const char *src, char *dest,
-   int allchars_mode)
+static void dict_make_global_alphabet (void)
 {
-   int c;
+   int i;
+   int j = 0;
+   int ch;
 
-   for (; *src; ++src) {
-      c = * (const unsigned char *) src;
+   for (i=0; i < charcount; ++i){
+      global_alphabet_8bit [i] = ch = c(i);
 
-      if (isspace( c )) {
-         *dest++ = ' ';
-      }else if (allchars_mode || isalnum( c )){
-	 *dest++ = tolower (c);
-      }
+      if (ch < 128)
+	 global_alphabet_ascii [j++] = ch;
    }
 
-   *dest = '\0';
-   return 1;
-}
+   global_alphabet_8bit  [charcount] = 0;
+   global_alphabet_ascii [j] = 0;
 
-/*
-  Copies alphanumeric and space characters converting them to lower case.
-  Strings are represented in UTF-8 character set.
-*/
-static int tolower_alnumspace_utf8 (
-   const char *src, char *dest,
-   int allchars_mode)
-{
-   wchar_t      ucs4_char;
-   size_t len;
-   int    len2;
-
-   mbstate_t ps;
-   mbstate_t ps2;
-
-   memset (&ps,  0, sizeof (ps));
-   memset (&ps2, 0, sizeof (ps2));
-
-   while (src && src [0]){
-      len = mbrtowc (&ucs4_char, src, MB_CUR_MAX, &ps);
-      if ((int) len < 0)
-	 return 0;
-
-      if (iswspace (ucs4_char)){
-	 *dest++ = ' ';
-      }else if (allchars_mode || iswalnum (ucs4_char)){
-	 len2 = wcrtomb (dest, towlower (ucs4_char), &ps2);
-	 if (len2 < 0)
-	    return 0;
-
-	 dest += len2;
-      }
-
-      src += len;
-   }
-
-   *dest = 0;
-
-   return (src != NULL);
+   PRINTF(DBG_SEARCH, (
+	  "global_alphabet_8bit  = '%s'\n",
+	  global_alphabet_8bit));
+   PRINTF(DBG_SEARCH, (
+	  "global_alphabet_ascii = '%s'\n",
+	  global_alphabet_ascii));
 }
 
 static void dict_table_init(void)
@@ -231,6 +203,9 @@ static void dict_table_init(void)
 		c2i(i), c2i(i),
 		i2c(c2i(i)), (char) i2c(c2i(i)) ? i2c(c2i(i)) : '.');
    }
+
+
+   dict_make_global_alphabet ();
 }
 
 static int compare_allchars(
@@ -326,9 +301,16 @@ static int compare_alnumspace(
 	 }
 	 if (dbg_test(DBG_SEARCH)){
 	    if (utf8_mode)
-	       printf("   result = %d (%i != %i) \n", result, c1, c2);
+	       printf(
+		  "   result = %d (%i != %i) \n", result, c1, c2);
 	    else
-	       printf("   result = %d ('%c' != '%c') \n", result, c1, c2);
+	       printf(
+		  "   result = %d ('%c'(c2i=%i) != '%c'(c2i=%i)) \n",
+		  result,
+		  c1,
+		  c2i (c1),
+		  c2,
+		  c2i (c2));
 	 }
          return result;
       }
@@ -382,8 +364,8 @@ static int compare(
       }
 
       *d = '\0';
-      printf( "compare \"%s\" with \"%s\" (sizes: %i and %i)\n",
-         word, buf, strlen( word ), strlen( buf ) );
+      printf( "compare \"%s\" with \"%s\" (sizes: %lu and %lu)\n",
+         word, buf, (unsigned long) strlen( word ), (unsigned long) strlen( buf ) );
    }
 
    ++_dict_comparisons;		/* counter for profiling */
@@ -441,7 +423,7 @@ static const char *binary_search_8bit(
 
    assert (dbindex);
 
-   PRINTF(DBG_SEARCH,("%s %p %p\n",word,start,end));
+   PRINTF(DBG_SEARCH,("word/start/end %s/%p/%p\n",word,start,end));
 
    pt = start + (end-start)/2;
    FIND_NEXT(pt,end);
@@ -455,8 +437,8 @@ static const char *binary_search_8bit(
 	 }
 
          *d = '\0';
-         printf( "compare \"%s\" with \"%s\" (sizes: %i and %i)\n",
-            word, buf, strlen( word ), strlen( buf ) );
+         printf( "compare \"%s\" with \"%s\" (sizes: %lu and %lu)\n",
+            word, buf, (unsigned long) strlen( word ), (unsigned long) strlen( buf ) );
       }
 
       if (
@@ -510,13 +492,11 @@ static const char *linear_search(
    return NULL;
 }
 
-const char *dict_index_search( const char *word, dictIndex *idx )
+static const char *dict_index_search( const char *word, dictIndex *idx )
 {
    const char    *start;
    const char    *end;
-#if OPTSTART
    int first, last;
-#endif
 
    assert (idx);
 
@@ -533,23 +513,24 @@ const char *dict_index_search( const char *word, dictIndex *idx )
       2097035 comparisons, 2000 words
    */
 
-#if OPTSTART
-   first   = * (const unsigned char *) word;
-   last    = i2c(c2i(first)+1);
+   if (optStart_mode){
+      first   = * (const unsigned char *) word;
+      last    = i2c(c2i(first)+1);
 
-   if (dbg_test(DBG_SEARCH)) {
-      if (!utf8_mode || (last <= CHAR_MAX && first <= CHAR_MAX))
-	 printf("binary_search from %c to %c\n", first, last);
-      else
-	 printf("binary_search from %i to %i\n", first, last);
+      if (dbg_test(DBG_SEARCH)) {
+	 if (!utf8_mode || (last <= CHAR_MAX && first <= CHAR_MAX))
+	    printf("binary_search from %c to %c\n", first, last);
+	 else
+	    printf("binary_search from %i to %i\n", first, last);
+      }
+
+      end   = idx->optStart [last];
+      start = idx->optStart [first];
+   }else{
+      start = idx->start;
+      end   = idx->end;
    }
 
-   end   = idx->optStart [last];
-   start = idx->optStart [first];
-#else
-   start = idx->start;
-   end   = idx->end;
-#endif
    if (end < start) end = idx->end;
 
    start = binary_search( word, idx, start, end );
@@ -602,8 +583,8 @@ static dictWord *dict_word_create(
 		    "Too few tabs in index entry \"%20.20s\"\n", entry );
 
    buf = alloca( newline + 1 );
-   strncpy( buf, entry, newline );
-   buf[firstTab] = buf[secondTab] = buf[newline] = '\0';
+   memcpy (buf, entry, newline);
+   buf[firstTab] = buf[secondTab] = buf [newline] = '\0';
 
    dw->start    = b64_decode( buf + firstTab + 1 );
    dw->end      = b64_decode( buf + secondTab + 1 );
@@ -999,6 +980,25 @@ static int dict_search_word(
    }
 }
 
+/* return non-zero if success, 0 otherwise */
+static int dict_match (
+   const regex_t *re,
+   const char *word, size_t word_len,
+   int eflags)
+{
+#if defined(REG_STARTEND)
+   regmatch_t    subs[1];
+   subs [0].rm_so = 0;
+   subs [0].rm_eo = word_len;
+   return !regexec(re, word, 1, subs, eflags | REG_STARTEND);
+#else
+   char *word_copy = (char *) alloca (word_len + 1);
+   memcpy (word_copy, word, word_len);
+   word_copy [word_len] = 0;
+   return !regexec(re, word_copy, 0, NULL, eflags);
+#endif
+}
+
 static int dict_search_regexpr( lst_List l,
 				const char *word,
 				const dictDatabase *database,
@@ -1013,20 +1013,26 @@ static int dict_search_regexpr( lst_List l,
    regex_t       re;
    char          erbuf[100];
    int           err;
-   regmatch_t    subs[1];
    unsigned char first;
    const char    *previous = NULL;
 
    assert (dbindex);
 
-#if OPTSTART
-   if (*word == '^' && dbindex -> isspacealnum [(unsigned char) word[1]]) {
-      first = word[1];
-      end   = dbindex->optStart[i2c(c2i(first)+1)];
-      start = dbindex->optStart[first];
-      if (end < start) end = dbindex->end;
+   if (optStart_mode){
+      if (
+	 *word == '^'
+	 && dbindex -> isspacealnum [(unsigned char) word[1]]
+	 && strchr (word, '|') == NULL)
+      {
+	 first = word[1];
+
+	 end   = dbindex->optStart[i2c(c2i(first)+1)];
+	 start = dbindex->optStart[first];
+
+	 if (end < start)
+	    end = dbindex->end;
+      }
    }
-#endif
 
    if ((err = regcomp(&re, word, REG_ICASE|REG_NOSUB|type))) {
       regerror(err, &re, erbuf, sizeof(erbuf));
@@ -1037,10 +1043,9 @@ static int dict_search_regexpr( lst_List l,
    pt = start;
    while (pt && pt < end) {
       for (p = pt; *p != '\t' && p < end; ++p);
-      subs[0].rm_so = 0;
-      subs[0].rm_eo = p - pt;
       ++_dict_comparisons;
-      if (!regexec(&re, pt, 1, subs, REG_STARTEND)) {
+
+      if (dict_match (&re, pt, p - pt, 0)) {
 	 if (!previous || altcompare(previous, pt, end)) {
 	    ++count;
 	    datum = dict_word_create( previous = pt, database, dbindex );
@@ -1074,7 +1079,7 @@ static int dict_search_regexp( lst_List l,
 			       const dictDatabase *database,
 			       dictIndex    *dbindex)
 {
-   return dict_search_regexpr( l, word, database, dbindex, REG_BASIC );
+   return dict_search_regexpr( l, word, database, dbindex, 0 /*REG_BASIC*/ );
 }
 
 static int dict_search_soundex( lst_List l,
@@ -1086,7 +1091,8 @@ static int dict_search_soundex( lst_List l,
    const char *end;
    int        count = 0;
    dictWord   *datum;
-   char       soundex[10];
+   char       soundex  [10];
+   char       soundex2 [5];
    char       buffer[MAXWORDLEN];
    char       *d;
    const unsigned char *s;
@@ -1096,17 +1102,18 @@ static int dict_search_soundex( lst_List l,
 
    assert (dbindex);
 
-#if OPTSTART
-   pt  = dbindex->optStart[ c ];
-   end = dbindex->optStart[ i2c(c2i(c)+1) ];
-   if (end < pt) end = dbindex->end;
-#else
-   pt  = dbindex->start;
-   end = dbindex->end;
-#endif
+   if (optStart_mode){
+      pt  = dbindex->optStart[ c ];
+      end = dbindex->optStart[ i2c(c2i(c)+1) ];
+      if (end < pt)
+	 end = dbindex->end;
+   }else{
+      pt  = dbindex->start;
+      end = dbindex->end;
+   }
+
+   txt_soundex2 (word, soundex);
 
-   strcpy( soundex, txt_soundex( word ) );
-   
    while (pt && pt < end) {
       for (i = 0, s = pt, d = buffer; i < MAXWORDLEN - 1; i++, ++s) {
 	 if (*s == '\t') break;
@@ -1114,7 +1121,9 @@ static int dict_search_soundex( lst_List l,
 	 *d++ = *s;
       }
       *d = '\0';
-      if (!strcmp(soundex, txt_soundex(buffer))) {
+
+      txt_soundex2 (buffer, soundex2);
+      if (!strcmp (soundex, soundex2)) {
 	 if (!previous || altcompare(previous, pt, end)) {
 	    datum = dict_word_create( previous = pt, database, dbindex );
 	    lst_append( l, datum );
@@ -1143,86 +1152,53 @@ static int dict_search_soundex( lst_List l,
 
 */
 
-static int dict_search_levenshtein( lst_List l,
-				    const char *word,
-				    const dictDatabase *database,
-				    dictIndex *dbindex)
-{
-   int        len   = strlen(word);
-   char       *buf  = alloca(len+2);
-   char       *p    = buf;
-   int        count = 0;
-   set_Set    s     = set_create(NULL,NULL);
-   int        i, j, k;
-   const char *pt;
-   char       tmp;
-   dictWord   *datum;
+typedef struct lev_args_ {
+   const dictDatabase *database;
+   dictIndex          *dbindex;
+   lst_List            l;
+} LEV_ARGS;
 
-   assert (dbindex);
-
-#define CHECK                                         \
-   if ((pt = dict_index_search(buf, dbindex))           \
-       && !compare(buf, dbindex, pt, dbindex->end)) {            \
-      if (!set_member(s,buf)) {                       \
-	 ++count;                                     \
-	 set_insert(s,str_find(buf));                 \
-	 datum = dict_word_create(pt, database, dbindex);\
-	 lst_append(l, datum);                        \
-         PRINTF(DBG_LEV,("  %s added\n",buf));        \
+#define CHECK(word, args)                                \
+   if ((pt = dict_index_search((word), (args) -> dbindex))         \
+       && !compare((word), (args) -> dbindex, pt, (args) -> dbindex -> end)) \
+   { \
+      if (!set_member(s,(word))) {                       \
+	 ++count;                                        \
+	 set_insert(s,str_find((word)));                 \
+	 datum = dict_word_create(pt, (args) -> database, (args) -> dbindex);\
+	 lst_append((args) -> l, datum);                        \
+         PRINTF(DBG_LEV,("  %s added\n",(word)));     \
       }                                               \
    }
 
-				/* Deletions */
-   for (i = 0; i < len; i++) {
-      p = buf;
-      for (j = 0; j < len; j++)
-	 if (i != j) *p++ = word[j];
-      *p = '\0';
-      CHECK;
-   }
-                                /* Transpositions */
-   for (i = 1; i < len; i++) {
-      strcpy( buf, word );
-      tmp = buf[i-1];
-      buf[i-1] = buf[i];
-      buf[i] = tmp;
-      CHECK;
-   }
+#define LEV_VARS \
+      char tmp;
 
-				/* Insertions */
-   for (i = 0; i < len; i++) {
-      for (k = 0; k < charcount; k++) {
-	 p = buf;
-         for (j = 0; j < len; j++) {
-            *p++ = word[j];
-            if (i == j) *p++ = c(k);
-         }
-         *p = '\0';
-	 CHECK;
+#include "lev.h"
+
+static int dict_search_levenshtein (
+   lst_List l,
+   const char *word,
+   const dictDatabase *database,
+   dictIndex *dbindex)
+{
+   LEV_ARGS lev_args = { database, dbindex, l };
+
+   assert (database);
+   assert (dbindex);
+
+   if (database -> alphabet){
+      return dict_search_lev (
+	 word, database -> alphabet, dbindex -> flag_utf8, &lev_args);
+   }else{
+      if (dbindex -> flag_utf8){
+	 return dict_search_lev (
+	    word, global_alphabet_ascii, 1, &lev_args);
+      }else{
+	 return dict_search_lev (
+	    word, global_alphabet_8bit, 0, &lev_args);
       }
    }
-                                /* Insertions at the end */
-   strcpy( buf, word );
-   buf[ len + 1 ] = '\0';
-   for (k = 0; k < charcount; k++) {
-      buf[ len ] = c(k);
-      CHECK;
-   }
-
-   
-                                  /* Substitutions */
-   for (i = 0; i < len; i++) {
-      strcpy( buf, word );
-      for (j = 0; j < charcount; j++) {
-         buf[i] = c(j);
-	 CHECK;
-      }
-   }
-
-   PRINTF(DBG_LEV,("  Got %d matches\n",count));
-   set_destroy(s);
-   
-   return count;
 }
 
 /*
@@ -1252,6 +1228,7 @@ static void stranagram_8bit (char *str, int length)
    }
 }
 
+#if HAVE_UTF8
 /*
   makes anagram of the utf-8 string 'str'
   Returns non-zero if success, 0 otherwise
@@ -1281,18 +1258,24 @@ static int stranagram_utf8 (char *str)
    stranagram_8bit (str, -1);
    return 1;
 }
+#endif
 
 /* makes anagram of utf-8 string 'str' */
 static int stranagram (char *str, int utf8_string)
 {
    assert (str);
 
+#if HAVE_UTF8
    if (utf8_string){
       return stranagram_utf8 (str);
    }else{
       stranagram_8bit (str, -1);
       return 1;
    }
+#else
+   stranagram_8bit (str, -1);
+   return 1;
+#endif
 }
 
 static int dict_search_suffix(
@@ -1334,144 +1317,54 @@ static int dict_search_suffix(
    }
 }
 
-#ifdef USE_PLUGIN
-static int dict_search_plugin (
-   lst_List l,
-   const char *const word,
-   const dictDatabase *database,
-   int strategy,
-   int *extra_result,
-   const dictPluginData **extra_data,
-   int *extra_data_size)
-{
-   int ret;
-   int                  failed = 0;
-   const char * const * defs;
-   const int          * defs_sizes;
-   int                  defs_count;
-   const char         * err_msg;
-   int                  i;
-   dictWord           * def;
-   int                  len;
-
-   assert (database);
-   assert (database -> plugin);
-
-   PRINTF (DBG_SEARCH, (":S:     searching\n"));
-
-   failed = database -> plugin -> dictdb_search (
-      database -> plugin -> data,
-      word, -1,
-      strategy,
-      &ret,
-      extra_data, extra_data_size,
-      &defs, &defs_sizes, &defs_count);
-
-   if (extra_result)
-      *extra_result = ret;
-
-   if (failed){
-      err_msg = database -> plugin -> dictdb_error (
-	 database -> plugin -> data);
-
-      PRINTF (DBG_SEARCH, (":E: Plugin failed: %s\n", err_msg ? err_msg : ""));
-   }else{
-      switch (ret){
-      case DICT_PLUGIN_RESULT_FOUND:
-	 PRINTF (DBG_SEARCH, (":S:     found %i definitions\n", defs_count));
-	 break;
-      case DICT_PLUGIN_RESULT_NOTFOUND:
-	 PRINTF (DBG_SEARCH, (":S:     not found\n"));
-	 return 0;
-      case DICT_PLUGIN_RESULT_EXIT:
-	 PRINTF (DBG_SEARCH, (":S:     exiting\n"));
-	 return 0;
-      case DICT_PLUGIN_RESULT_PREPROCESS:
-	 PRINTF (DBG_SEARCH, (":S:     preprocessing\n"));
-	 break;
-      default:
-	 err_fatal (__FUNCTION__, "invalid pligin's exit status\n");
-      }
-
-      for (i = 0; i < defs_count; ++i){
-	 def = xmalloc (sizeof (dictWord));
-	 memset (def, 0, sizeof (*def));
-
-	 def -> database = database;
-	 def -> start    = def -> end = 0;
-
-	 len = defs_sizes [i];
-	 if (-1 == len)
-	    len = strlen (defs [i]);
-
-	 if (
-	    strategy & DICT_MATCH_MASK &&
-	    ret != DICT_PLUGIN_RESULT_PREPROCESS)
-	 {
-	    def -> word     = xstrdup (defs [i]);
-	    def -> def      = def -> word;
-	    def -> def_size = -1;
-	 }else{
-	    def -> word     = xstrdup (word);
-	    def -> def      = defs [i];
-	    def -> def_size = len;
-	 }
-
-	 lst_push (l, def);
-      }
-
-      return defs_count;
-   }
-
-   return 0;
-}
-#endif /* USE_PLUGIN */
-
+#if HAVE_UTF8
 static const char *utf8_err_msg = "\
 error: The request is not a valid UTF-8 string";
+#endif
 
 /*
   returns a number of matches ( >= 0 ) or
   negative value for invalid UTF-8 string
 */
-static int dict_search_database_ (
+int dict_search_database_ (
    lst_List l,
    const char *const word,
    const dictDatabase *database,
    int strategy )
 {
    char       *buf      = NULL;
+#if HAVE_UTF8
    dictWord   *dw       = NULL;
+#endif
 
    assert (database);
    assert (database -> index);
 
    buf = alloca( strlen( word ) + 1 );
 
-   if (utf8_mode){
-      if (
-	 !strcmp(utf8_err_msg, word) ||
-	 !tolower_alnumspace_utf8 (
-	    word, buf, database -> index -> flag_allchars))
-      {
-	 PRINTF(DBG_SEARCH, ("tolower_... ERROR!!!\n"));
-
-	 dw = xmalloc (sizeof (dictWord));
-	 memset (dw, 0, sizeof (dictWord));
-
-	 dw -> database = database;
-	 dw -> def      = utf8_err_msg;
-	 dw -> def_size = -1;
-	 dw -> word     = strdup (word);
-
-	 lst_append (l, dw);
-
-	 return -1;
-      }
-   }else{
-      tolower_alnumspace_8bit (
-	  word, buf, database -> index -> flag_allchars);
+#if HAVE_UTF8
+   if (
+      !strcmp(utf8_err_msg, word) ||
+      tolower_alnumspace (
+	 word, buf, database -> index -> flag_allchars, utf8_mode))
+   {
+      PRINTF(DBG_SEARCH, ("tolower_... ERROR!!!\n"));
+      
+      dw = xmalloc (sizeof (dictWord));
+      memset (dw, 0, sizeof (dictWord));
+      
+      dw -> database = database;
+      dw -> def      = utf8_err_msg;
+      dw -> def_size = -1;
+      dw -> word     = strdup (word);
+      
+      lst_append (l, dw);
+      
+      return -1;
    }
+#else
+   tolower_alnumspace (word, buf, database -> index -> flag_allchars, utf8_mode);
+#endif
 
    if (!buf [0] && word [0]){
       /*
@@ -1504,7 +1397,7 @@ static int dict_search_database_ (
       return dict_search_substring( l, buf, database, database->index );
 
    case DICT_SUFFIX:
-      return dict_search_suffix( l, word, database );
+      return dict_search_suffix( l, buf, database );
 
    case DICT_RE:
       return dict_search_re( l, word, database, database->index );
@@ -1529,192 +1422,30 @@ static int dict_search_database_ (
    }
 }
 
-#ifdef USE_PLUGIN
-/* reads data without headword 00-... */
-static char *dict_plugin_data (const dictDatabase *db, const dictWord *dw)
-{
-   char *buf = dict_data_obtain (db, dw);
-   char *p = buf;
-   int len;
-
-   assert (db);
-   assert (db -> index);
-
-   if (!strncmp (p, DICT_ENTRY_PLUGIN_DATA, strlen (DICT_ENTRY_PLUGIN_DATA))){
-      while (*p != '\n')
-	 ++p;
-   }
-
-   while (*p == '\n')
-      ++p;
-
-   len = strlen (p);
-
-   while (len > 0 && p [len - 1] == '\n')
-      --len;
-
-   p [len] = 0;
-
-   p = xstrdup (p);
-   xfree (buf);
-
-   return p;
-}
-
-static int plugin_initdata_set_data (
-   dictPluginData *data, int data_size,
+/*
+  Replaces invisible databases with db argument.
+ */
+static void replace_invisible_databases (
+   lst_Position pos,
    const dictDatabase *db)
 {
-   char *plugin_data;
-   int ret = 0;
-   lst_List list;
    dictWord *dw;
 
-   if (!db -> index)
-      return 0;
-
-   if (data_size <= 0)
-      err_fatal (__FUNCTION__, "invalid initial array size");
-
-   list = lst_create ();
-
-   ret = dict_search_database_ (
-      list, DICT_ENTRY_PLUGIN_DATA, db, DICT_EXACT);
-
-   if (0 == ret){
-      lst_destroy (list);
-      return 0;
-   }
-
-   dw = (dictWord *) lst_pop (list);
-   plugin_data = dict_plugin_data (db, dw);
-
-   dict_destroy_datum (dw);
-   if (2 == ret)
-      dict_destroy_datum (lst_pop (list));
-
-   data -> id   = DICT_PLUGIN_INITDATA_DICT;
-   data -> data = plugin_data;
-   data -> size = -1;
-
-   lst_destroy (list);
-
-   return 1;
-}
-
-static int plugin_initdata_set_dbnames (dictPluginData *data, int data_size)
-{
-   const dictDatabase *db;
-   int count;
-   int i;
-
-   if (data_size <= 0)
-      err_fatal (__FUNCTION__, "too small initial array");
-
-   count = lst_length (DictConfig -> dbl);
-   if (count == 0)
-      return 0;
-
-   if (count > data_size)
-      err_fatal (__FUNCTION__, "too small initial array");
-
-   for (i = 1; i <= count; ++i){
-      db = (const dictDatabase *)(lst_nth_get (DictConfig -> dbl, i));
-
-      data -> id   = DICT_PLUGIN_INITDATA_DBNAME;
-      if (db -> databaseName){
-	 data -> size = strlen (db -> databaseName);
-	 data -> data = xstrdup (db -> databaseName);
-      }else{
-	 data -> size = 0;
-	 data -> data = NULL;
-      }
-
-      ++data;
-   }
-
-   return count;
-}
-
-static int plugin_initdata_set_stratnames (dictPluginData *data, int data_size)
-{
-   dictStrategy **strats;
-   int count;
-   int ret = 0;
-   int i;
-   dictPluginData_strategy datum;
-
-   if (data_size <= 0)
-      err_fatal (__FUNCTION__, "too small initial array");
-
-   count = get_strategy_count ();
-   assert (count > 0);
-
-   strats = get_strategies ();
-
-   for (i = 0; i < count; ++i){
-      data -> id   = DICT_PLUGIN_INITDATA_STRATEGY;
+   while (pos){
+      dw = (dictWord *) lst_get_position (pos);
 
       if (
-	 strlen (strats [i] -> name) + 1 >
-	 sizeof (datum.name))
+	 dw -> database &&
+	 dw -> database -> invisible &&
+	 !dw -> database_visible)
       {
-	 err_fatal (__FUNCTION__, "too small initial array");
+	 dw -> database_visible = db;
       }
 
-      datum.number = strats [i] -> number;
-      strcpy (datum.name, strats [i] -> name);
-
-      data -> size = sizeof (datum);
-      data -> data = xmalloc (sizeof (datum));
-
-      memcpy ((void *) data -> data, &datum, sizeof (datum));
-
-      ++data;
-      ++ret;
-   }
-
-   return ret;
-}
-
-/* all dict [i]->data are xmalloc'd*/
-static int plugin_initdata_set (
-   dictPluginData *data, int data_size,
-   const dictDatabase *db)
-{
-   int count = 0;
-   dictPluginData *p = data;
-
-   count = plugin_initdata_set_dbnames (data, data_size);
-   data      += count;
-   data_size -= count;
-
-   count = plugin_initdata_set_stratnames (data, data_size);
-   data      += count;
-   data_size -= count;
-
-   count = plugin_initdata_set_data (data, data_size, db);
-   data      += count;
-   data_size -= count;
-
-   return data - p;
-}
-
-static void plugin_init_data_free (
-   dictPluginData *data, int data_size)
-{
-   int i=0;
-
-   for (i = 0; i < data_size; ++i){
-      if (data -> data)
-	 xfree ((void *) data -> data);
-
-      ++data;
+      pos = lst_next_position (pos);
    }
 }
 
-#endif /* USE_PLUGIN */
-
 /*
   returns a number of matches ( >= 0 ) or
   negative value for invalid UTF-8 string
@@ -1729,26 +1460,29 @@ int dict_search (
    int *extra_data_size)
 {
    int count = 0;
-#ifdef USE_PLUGIN
-   int res = 0;
-#endif
    dictWord *dw;
 
+   int norm_strategy = strategy & ~DICT_MATCH_MASK;
+
+   if (extra_result)
+      *extra_result = DICT_PLUGIN_RESULT_NOTFOUND;
+
    assert (word);
    assert (database);
 
-   if (!database -> index && !strcmp (word, DICT_INFO_ENTRY_NAME)){
-      dw = xmalloc (sizeof (dictWord));
-      memset (dw, 0, sizeof (dictWord));
-
-      dw -> database = database;
-      dw -> word     = strdup (word);
-      dw -> def      = database -> databaseShort;
-      dw -> def_size = -1;
-      lst_append (l, dw);
-      count = 1;
-
-   }else{
+   if (
+      database -> strategy_disabled &&
+      database -> strategy_disabled [norm_strategy])
+   {
+      /* disable_strategy keyword from configuration file */
+#if 0
+      PRINTF (DBG_SEARCH, (
+	 ":S: strategy '%s' is disabled for database '%s'\n",
+	 get_strategies () [norm_strategy] -> name,
+	 database -> databaseName ? database -> databaseName : "(unknown)"));
+#endif
+      return 0;
+   }
 
       PRINTF (DBG_SEARCH, (":S: Searching in '%s'\n", database -> databaseName));
 
@@ -1756,100 +1490,49 @@ int dict_search (
       fprintf (stderr, "STRATEGY: %x\n", strategy);
 #endif
 
+      if (database -> index){
+	 PRINTF (DBG_SEARCH, (":S:   database search\n"));
+	 count = dict_search_database_ (l, word, database, norm_strategy);
+      }
+
 #ifdef USE_PLUGIN
-      if (database -> plugin){
+      if (!count && database -> plugin){
 	 PRINTF (DBG_SEARCH, (":S:   plugin search\n"));
 	 count = dict_search_plugin (
 	    l, word, database, strategy,
-	    &res, extra_data, extra_data_size);
-
-	 if (extra_result)
-	    *extra_result = res;
+	    extra_result, extra_data, extra_data_size);
 
 	 if (count)
 	    return count;
-
-	 switch (res){
-	 case DICT_PLUGIN_RESULT_EXIT:
-	    return 0;
-	 case DICT_PLUGIN_RESULT_NOTFOUND:
-	    if (strncmp (word, DICT_ENTRY_PREFIX, DICT_ENTRY_PREFIX_LEN))
-	       return 0;
-	 default:
-	    break;
-	 }
       }
 #endif
 
-      if (database -> index){
-	 strategy &= ~DICT_MATCH_MASK;
+      if (!count && database -> virtual_db_list){
+	 lst_Position db_list_pos;
+	 dictDatabase *db = NULL;
+	 int old_count = lst_length (l);
 
-	 PRINTF (DBG_SEARCH, (":S:   database search\n"));
-	 count = dict_search_database_ (l, word, database, strategy);
-      }
-   }
+	 assert (lst_init_position (database -> virtual_db_list));
 
-   if (extra_result){
-      if (count != 0){
-	 *extra_result = DICT_PLUGIN_RESULT_FOUND;
-      }else{
-	 *extra_result = DICT_PLUGIN_RESULT_NOTFOUND;
+	 LST_ITERATE (database -> virtual_db_list, db_list_pos, db){
+	    count += dict_search (
+	       l, word, db, strategy,
+	       extra_result, extra_data, extra_data_size);
+	 }
+
+	 if (count > 0){
+	    replace_invisible_databases (
+	       lst_nth_position (l, old_count + 1),
+	       database);
+	 }
       }
-   }
+
+   if (count > 0 && extra_result)
+      *extra_result = DICT_PLUGIN_RESULT_FOUND;
 
    return count;
 }
 
-#ifdef USE_PLUGIN
-/* Reads plugin's file name from .dict file */
-/* do not free() returned value*/
-static char *dict_plugin_filename (
-   const dictDatabase *db,
-   const dictWord *dw)
-{
-   static char filename [FILENAME_MAX];
-
-   char *buf = dict_data_obtain (db, dw);
-   char *p = buf;
-   int len;
-
-   if (!strncmp (p, DICT_ENTRY_PLUGIN, strlen (DICT_ENTRY_PLUGIN))){
-      while (*p != '\n')
-	 ++p;
-   }
-
-   while (*p == '\n' || isspace ((unsigned char) *p))
-      ++p;
-
-   len = strlen (p);
-
-   while (
-      len > 0 &&
-      (p [len - 1] == '\n' || isspace ((unsigned char) p [len - 1])))
-   {
-      --len;
-   }
-
-   p [len] = 0;
-
-   if (p [0] != '.' && p [0] != '/'){
-      if (sizeof (filename) < strlen (DICT_PLUGIN_PATH) + strlen (p) + 1)
-	 err_fatal (__FUNCTION__, "too small initial array\n");
-
-      strcpy (filename, DICT_PLUGIN_PATH);
-      strcat (filename, p);
-   }else{
-      strncpy (filename, p, sizeof (filename) - 1);
-   }
-
-   filename [sizeof (filename) - 1] = 0;
-
-   xfree (buf);
-
-   return filename;
-}
-#endif /* USE_PLUGIN */
-
 dictIndex *dict_index_open(
    const char *filename,
    int init_flags, int flag_utf8, int flag_allchars)
@@ -1858,10 +1541,11 @@ dictIndex *dict_index_open(
    static int  tabInit = 0;
    dictIndex   *i;
    dictDatabase db;
-#if OPTSTART
    int         j;
    char        buf[2];
-#endif
+
+   int first_char;
+   int first_char_uc;
 
    if (!filename)
       return NULL;
@@ -1911,10 +1595,10 @@ dictIndex *dict_index_open(
    i->flag_allchars = flag_allchars;
    i->isspacealnum  = isspacealnumtab;
 
-#if OPTSTART
-   for (j = 0; j <= UCHAR_MAX; j++)
-      i->optStart[j] = i->start;
-#endif
+   if (optStart_mode){
+      for (j = 0; j <= UCHAR_MAX; j++)
+	 i->optStart[j] = i->start;
+   }
 
    if (init_flags){
       memset (&db, 0, sizeof (db));
@@ -1935,7 +1619,8 @@ dictIndex *dict_index_open(
 	 0 != dict_search_database_ (NULL, DICT_FLAG_UTF8, &db, DICT_EXACT);
       PRINTF(DBG_INIT, (":I:     \"%s\": flag_utf8=%i\n", filename, i->flag_utf8));
       if (i->flag_utf8 && !utf8_mode){
-	 fprintf (stderr, "\"C\" locale can not be used for utf-8 dictionaries\n");
+	 fprintf (stderr, "locale '%s' can not be used for utf-8 dictionaries\n", locale);
+	 log_info( ":E: locale '%s' can not be used for utf-8 dictionaries. Exiting\n", locale );
 	 exit (1);
       }
 
@@ -1944,39 +1629,57 @@ dictIndex *dict_index_open(
 	 0 != dict_search_database_ (NULL, DICT_FLAG_8BIT, &db, DICT_EXACT);
       PRINTF(DBG_INIT, (":I:     \"%s\": flag_8bit=%i\n", filename, i->flag_8bit));
       if (i->flag_8bit && !bit8_mode){
-	 fprintf (stderr, "\"C\" locale can not be used for 8-bit dictionaries\n");
+	 fprintf (stderr, "locale '%s' can not be used for 8-bit dictionaries\n", locale);
+	 log_info( ":E: locale '%s' can not be used for 8-bit dictionaries. Exiting\n", locale );
 	 exit (1);
       }
    }
 
-#if OPTSTART
-   buf[0] = ' ';
-   buf[1] = '\0';
-   i->optStart[ ' ' ] = binary_search_8bit( buf, i, i->start, i->end );
-
-   for (j = 0; j < charcount; j++) {
-      buf[0] = c(j);
+   if (optStart_mode){
+      buf[0] = ' ';
       buf[1] = '\0';
-      i->optStart[toupper(c(j))]
-	 = i->optStart[c(j)]
-	 = binary_search_8bit( buf, i, i->start, i->end );
-      if (dbg_test (DBG_SEARCH)){
-	 if (!utf8_mode || c(j) <= CHAR_MAX)
-	    printf ("optStart [%c] = %p\n", c(j), i->optStart[c(j)]);
-	 else
-	    printf ("optStart [%i] = %p\n", c(j), i->optStart[c(j)]);
+      i->optStart[ ' ' ] = binary_search_8bit( buf, i, i->start, i->end );
+
+      for (j = 0; j < charcount; j++) {
+	 first_char = c(j);
+
+	 buf[0] = first_char;
+	 buf[1] = '\0';
+
+	 i->optStart [first_char]
+	    = binary_search_8bit( buf, i, i->start, i->end );
+
+	 if (!utf8_mode || first_char < 128){
+	    first_char_uc = toupper (first_char);
+
+	    assert (first_char_uc >= 0 && first_char_uc <= UCHAR_MAX);
+
+	    i->optStart [first_char_uc] = i->optStart [first_char];
+	 }
+
+	 if (dbg_test (DBG_SEARCH)){
+	    if (!utf8_mode || first_char <= CHAR_MAX)
+	       printf (
+		  "optStart [%c] = %p\n",
+		  first_char,
+		  i->optStart [first_char]);
+	    else
+	       printf (
+		  "optStart [%i] = %p\n",
+		  first_char,
+		  i->optStart [first_char]);
+	 }
       }
-   }
 
-   for (j = '0'; j <= '9'; j++) {
-      buf[0] = j;
-      buf[1] = '\0';
-      i->optStart[j] = binary_search_8bit( buf, i, i->start, i->end );
-   }
+      for (j = '0'; j <= '9'; j++) {
+	 buf[0] = j;
+	 buf[1] = '\0';
+	 i->optStart[j] = binary_search_8bit( buf, i, i->start, i->end );
+      }
 
-   i->optStart[UCHAR_MAX]   = i->end;
-   i->optStart[UCHAR_MAX+1] = i->end;
-#endif
+      i->optStart[UCHAR_MAX]   = i->end;
+      i->optStart[UCHAR_MAX+1] = i->end;
+   }
 
    return i;
 }
@@ -2009,183 +1712,3 @@ void dict_index_close( dictIndex *i )
 
    xfree (i);
 }
-
-#ifdef USE_PLUGIN
-
-static void dict_plugin_test (dictPlugin *plugin, int version, int ret)
-{
-   const char *err_msg = NULL;
-
-   if (ret){
-      err_msg = plugin -> dictdb_error (
-	 plugin -> data);
-
-      if (err_msg){
-	 err_fatal (
-	    __FUNCTION__,
-	    "%s\n",
-	    plugin -> dictdb_error (plugin -> data));
-      }else{
-	 err_fatal (
-	    __FUNCTION__,
-	    "Error code %i\n", ret);
-      }
-   }
-
-   switch (version){
-   case 0:
-      break;
-/*
-   case 1:
-      if (!i -> plugin -> dictdb_set)
-	 err_fatal (__FUNCTION__, "'%s' function is not found\n", DICT_PLUGINFUN_SET);
-      break;
-*/
-   default:
-      err_fatal (__FUNCTION__, "Invalid version returned by plugin\n");
-   }
-}
-
-static void dict_plugin_dlsym (dictPlugin *plugin)
-{
-   PRINTF(DBG_INIT, (":I:     getting functions addresses\n"));
-
-   plugin -> dictdb_open   =
-      lt_dlsym (plugin -> handle, DICT_PLUGINFUN_OPEN);
-   plugin -> dictdb_free   =
-      lt_dlsym (plugin -> handle, DICT_PLUGINFUN_FREE);
-   plugin -> dictdb_search =
-      lt_dlsym (plugin -> handle, DICT_PLUGINFUN_SEARCH);
-   plugin -> dictdb_close  =
-      lt_dlsym (plugin -> handle, DICT_PLUGINFUN_CLOSE);
-   plugin -> dictdb_error  =
-      lt_dlsym (plugin -> handle, DICT_PLUGINFUN_ERROR);
-   plugin -> dictdb_set   =
-      lt_dlsym (plugin -> handle, DICT_PLUGINFUN_SET);
-
-   if (!plugin -> dictdb_open ||
-       !plugin -> dictdb_search ||
-       !plugin -> dictdb_free ||
-       !plugin -> dictdb_error ||
-       !plugin -> dictdb_close)
-   {
-      PRINTF(DBG_INIT, (":I:     faild\n"));
-      exit (1);
-   }
-}
-
-static dictPlugin *create_plugin (
-   const char *plugin_filename,
-   const dictPluginData *plugin_init_data,
-   int plugin_init_data_size)
-{
-   dictPlugin *plugin;
-   int ret;
-   int version;
-
-   PRINTF(DBG_INIT, (":I:   Initializing plugin '%s'\n", plugin_filename));
-
-   plugin = xmalloc (sizeof (dictPlugin));
-   memset (plugin, 0, sizeof (dictPlugin));
-
-   PRINTF(DBG_INIT, (":I:     opening plugin\n"));
-   plugin -> handle = lt_dlopen (plugin_filename);
-   if (!plugin -> handle){
-      PRINTF(DBG_INIT, (":I:     faild\n"));
-      exit (1);
-   }
-
-   dict_plugin_dlsym (plugin);
-
-   PRINTF(DBG_INIT, (":I:     initializing plugin\n"));
-   ret = plugin -> dictdb_open (
-      plugin_init_data, plugin_init_data_size, &version, &plugin -> data);
-
-   dict_plugin_test (plugin, version, ret);
-
-   return plugin;
-}
-
-int dict_plugin_open (dictDatabase *db)
-{
-   int ret = 0;
-   lst_List list;
-   const char *plugin_filename;
-   dictWord *dw;
-
-   dictPluginData init_data [3000];
-   int init_data_size;
-
-   init_data_size = plugin_initdata_set (
-      init_data, sizeof (init_data)/sizeof (init_data [0]),
-      db);
-
-   if (db -> plugin_db){
-      init_data [init_data_size].id   = DICT_PLUGIN_INITDATA_DICT;
-      init_data [init_data_size].data = db -> plugin_data;
-      init_data [init_data_size].size = -1;
-
-      db -> plugin = create_plugin (
-	 db -> pluginFilename,
-	 init_data, init_data_size + 1);
-   }else{
-      if (db -> index){
-	 list = lst_create ();
-
-	 ret = dict_search_database_ (list, DICT_ENTRY_PLUGIN, db, DICT_EXACT);
-	 switch (ret){
-	 case 1: case 2:
-	    dw = (dictWord *) lst_pop (list);
-
-	    plugin_filename = dict_plugin_filename (db, dw);
-
-	    dict_destroy_datum (dw);
-	    if (2 == ret)
-	       dict_destroy_datum (lst_pop (list));
-
-	    db -> plugin = create_plugin (
-	       plugin_filename, init_data, init_data_size);
-
-	    break;
-	 case 0:
-	    break;
-	 default:
-	    err_internal( __FUNCTION__, "Corrupted .index file'\n" );
-	 }
-
-	 lst_destroy (list);
-      }
-   }
-
-   plugin_init_data_free (init_data, init_data_size);
-
-   return 0;
-}
-
-void dict_plugin_close ( dictDatabase *db )
-{
-   int ret;
-
-   if (!db)
-      return;
-
-   if (!db -> plugin)
-      return;
-
-   if (db -> plugin -> dictdb_close){
-      ret = db -> plugin -> dictdb_close (db -> plugin -> data);
-      if (ret){
-	 PRINTF(DBG_INIT, ("exiting plugin failed"));
-	 exit (1);
-      }
-   }
-
-   ret = lt_dlclose (db -> plugin -> handle);
-   if (ret)
-      PRINTF(DBG_INIT, ("%s", lt_dlerror ()));
-
-   xfree (db -> plugin);
-   db -> plugin = NULL;
-}
-
-#endif /* USE_PLUGIN */
